"""
define class PdBackground which describes the background in 1d powder diffraction experiment
"""
__author__ = 'ikibalin'
__version__ = "2019_09_06"
import os
import numpy


from pycifstar import Global
from cryspy.f_common.cl_fitable import Fitable


class PdBackground(object):
    """
    Example:

    loop_
    _pd_background_2theta
    _pd_background_intensity
     4.5  256.0
     40.0  158.0
     80.0  65.0
    """
    def __init__(self, ttheta=[], intensity=[]):
        super(PdBackground, self).__init__()
        self.__pd_background_2theta = []
        self.__pd_background_intensity = []
        self.ttheta = ttheta
        self.intensity = intensity

    @property
    def ttheta(self):
        return self.__pd_background_2theta
    @ttheta.setter
    def ttheta(self, l_x):
        l_x_in = []
        for x in l_x:
            x_in = float(x)
            l_x_in.append(x_in)
        self.__pd_background_2theta = l_x_in
        len_x = len(l_x_in)

        len_1 = len(self.intensity)
        if len_1 > len_x:
            self.__pd_background_intensity = self.__pd_background_intensity[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_pd_background_intensity") for hh in range(len_x-len_1)] #default
            self.__pd_background_intensity.extend(l_fitable)

    @property
    def intensity(self):
        return tuple(self.__pd_background_intensity)
    @intensity.setter
    def intensity(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                x_in.name = "_pd_background_intensity"
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.ttheta)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name="_pd_background_intensity") for hh in range(len_1-len_x)])
        self.__pd_background_intensity = l_fitable



            
    def __repr__(self):
        ls_out = ["PdBackground:"]
        ls_out.append(" ttheta intensity")
        for _1, _2 in zip(self.ttheta, self.intensity):
            ls_out.append("{:7.2f} {:}".format(_1, _2.print_with_sigma))
        return "\n".join(ls_out)

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_pd_background_2theta")
            ls_out.append("_pd_background_intensity")
            for _1, _2 in zip(self.ttheta, self.intensity):
                ls_out.append("{:} {:}".format(_1, _2.print_with_sigma))
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_pd_background")
        if flag:
            cif_loop = cif_global["_pd_background"]
            l_name = cif_loop.names
            if "_pd_background_2theta" in l_name:
                self.ttheta = [float(_1) for _1 in cif_loop["_pd_background_2theta"]]
            if "_pd_background_intensity" in l_name:
                self.intensity = [str(_1) for _1 in cif_loop["_pd_background_intensity"]]
        else:
            self.ttheta, self.intensity = [], []
        return True

    @property
    def is_defined(self):
        cond = all([self.ttheta is not None, self.intensity is not None])
        return cond

    @property
    def is_variable(self):
        res = any([_1.refinement for _1 in self.intensity])
        return res
    
    def get_variables(self):
        l_variable = []
        l_variable.extend([_1 for _1 in self.intensity if _1.refinement])
        return l_variable

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    def interpolate_by_points(self, tth):
        l_ttheta = self.ttheta
        l_intensity = self.intensity
        tth_b = numpy.array(l_ttheta, dtype=float)
        int_b = numpy.array(l_intensity, dtype=float)
        if len(l_ttheta) == 0:
            int_1d = numpy.zeros(tth.size, dtype=float)
        else:
            int_1d = numpy.interp(tth, tth_b, int_b)
        return int_1d



"""
define class Diffrn which describes the single diffraction experiment
"""
__author__ = 'ikibalin'
__version__ = "2019_09_09"
import os
import numpy
from pycifstar import Data, Loop

from cryspy.f_common.cl_fitable import Fitable
from cryspy.f_experiment.cl_beam_polarization import BeamPolarization
from cryspy.f_experiment.cl_orient_matrix import OrientMatrix
from cryspy.f_experiment.f_single.cl_diffrn_refln import DiffrnRefln
from cryspy.f_experiment.cl_extinction import Extinction



class Diffrn(object):
    """
    Class to describe information about single diffraction measurements
    """
    def __init__(self, label="mono", 
                 extinction = Extinction(), diffrn_refln=DiffrnRefln(), orient_matrix=OrientMatrix(),
                 beam_polarization=BeamPolarization(), 
                 wavelength=1.4, field=1.0, phase_label="phase1",
                 ):
        super(Diffrn, self).__init__()
        self.__label = None
        self.__extinction = None
        self.__diffrn_refln = None
        self.__orient_matrix = None
        self.__beam_polarization = None

        self.__diffrn_radiation_wavelength = None
        self.__diffrn_ambient_field = None
        self.__refln = None
        self.__phase_label = None

        self.label = label
        self.extinction = extinction
        self.diffrn_refln = diffrn_refln
        self.beam_polarization = beam_polarization
        self.wavelength = wavelength
        self.phase_label = phase_label
        
        orient_matrix.wavelength = wavelength
        self.orient_matrix = orient_matrix
        
        self.field = field

    @property
    def label(self):
        return self.__label
    @label.setter
    def label(self, x: str):
        self.__label = str(x)

    @property
    def phase_label(self):
        return self.__phase_label
    @phase_label.setter
    def phase_label(self, x: str):
        self.__phase_label = str(x)

    @property
    def extinction(self):
        return self.__extinction
    @extinction.setter
    def extinction(self, x):
        if isinstance(x, Extinction):
            x_in = x
        elif isinstance(x, str):
            x_in = Extinction()
            flag = x_in.from_cif(x)
            if not(flag):
                x_in = None
                self._show_message("A induced string can not be converted to Extinction")
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("Input type for extinction is not recognized")
        self.__extinction = x_in

    @property
    def diffrn_refln(self):
        return self.__diffrn_refln
    @diffrn_refln.setter
    def diffrn_refln(self, x):
        if isinstance(x, DiffrnRefln):
            x_in = x
        elif isinstance(x, str):
            x_in = DiffrnRefln()
            flag = x_in.from_cif(x)
            if not(flag):
                x_in = None
                self._show_message("A induced string can not be converted to DiffrnRefln")
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("Input type for diffrn_refln is not recognized")
        self.__diffrn_refln = x_in

    @property
    def orient_matrix(self):
        return self.__orient_matrix
    @orient_matrix.setter
    def orient_matrix(self, x):
        if isinstance(x, OrientMatrix):
            x_in = x
        elif isinstance(x, str):
            x_in = OrientMatrix()
            flag = x_in.from_cif(x)
            if not(flag):
                x_in = None
                self._show_message("A induced string can not be converted to OrientMatrix")
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("Input type for orient_matrix is not recognized")
        self.__orient_matrix = x_in

    @property
    def beam_polarization(self):
        return self.__beam_polarization
    @beam_polarization.setter
    def beam_polarization(self, x):
        if isinstance(x, BeamPolarization):
            x_in = x
        elif isinstance(x, str):
            x_in = BeamPolarization()
            flag = x_in.from_cif(x)
            if not(flag):
                x_in = None
                self._show_message("A induced string can not be converted to BeamPolarization")
        elif x is None:
            x_in = None
        else:
            x_in = None
            self._show_message("Input type for beam_polarization is not recognized")
        self.__beam_polarization = x_in

    @property
    def wavelength(self):
        return self.__diffrn_radiation_wavelength
    @wavelength.setter
    def wavelength(self, x):
        if isinstance(x, float):
            x_in = x
        elif x is None:
            x_in = None
        else:
            x_in = float(x)
        self.__diffrn_radiation_wavelength = x_in

    @property
    def field(self):
        return self.__diffrn_ambient_field
    @field.setter
    def field(self, x):
        if isinstance(x, float):
            x_in = x
        elif x is None:
            x_in = None
        else:
            x_in = float(x)
        self.__diffrn_ambient_field = x_in

    @property
    def refln(self):
        return self.__refln
    @refln.setter
    def refln(self, x):
        self.__refln = x


    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    def __repr__(self):
        ls_out = ["Diffrn"]
        if self.label is not None:
            ls_out.append("label: {:}".format(self.label))
        if self.wavelength is not None:
            ls_out.append("\nwavelength: {:.3f}".format(float(self.wavelength)))
        if self.field is not None:
            ls_out.append("field: {:.3f}".format(float(self.field)))
        if self.extinction is not None:
            ls_out.append("\n"+str(self.extinction))
        if self.beam_polarization is not None:
            ls_out.append("\n"+str(self.beam_polarization))
        if self.orient_matrix is not None:
            ls_out.append("\n"+str(self.orient_matrix))
        if self.diffrn_refln is not None:
            ls_out.append("\n"+str(self.diffrn_refln))
        if self.refln is not None:
            ls_out.append("\n"+str(self.refln))
        return "\n".join(ls_out)

    
    def calc_iint_u_d_flip_ratio(self, h, k, l, l_crystal):
        """
        calculate intensity for the given diffraction angle
        """
        l_label = [_.label for _ in l_crystal]
        if self.phase_label in l_label:
            ind = l_label.index(self.phase_label)
            crystal = l_crystal[ind]
        elif self.label in l_label:
            self.phase_label = self.label
            ind = l_label.index(self.label)
            crystal = l_crystal[ind]
        else:
            self._show_message("Crystal not found.\nThe first one is taken.")
            crystal = l_crystal[0]
            self.phase_label = crystal.label
        wavelength = self.wavelength
        field_z = self.field
        beam_polarization = self.beam_polarization
        extinction = self.extinction
        orient_matrix = self.orient_matrix

        orientation = orient_matrix.u 
        
        field_vec = numpy.array([0., 0., field_z], dtype=float)

        phi_d, chi_d, omega_d = 0., 0., 0.
        m_phi_d = numpy.array([[ numpy.cos(phi_d), numpy.sin(phi_d),0.],
                               [-numpy.sin(phi_d), numpy.cos(phi_d),0.],
                               [         0.,       0.,    1.]], dtype=float)

        m_omega_d = numpy.array([[ numpy.cos(omega_d), numpy.sin(omega_d),0.],
                                 [-numpy.sin(omega_d), numpy.cos(omega_d),0.],
                                 [       0.,       0.,    1.]], dtype=float)

        m_chi_d = numpy.array([[ numpy.cos(chi_d), 0., numpy.sin(chi_d)],
                               [               0., 1.,               0.],
                               [-numpy.sin(chi_d), 0., numpy.cos(chi_d)]], dtype=float)
        
        m_u_d = numpy.matmul(m_omega_d, numpy.matmul(m_chi_d, numpy.matmul(m_phi_d, 
                             orientation)))

        field_loc = numpy.matmul(m_u_d.transpose(), field_vec)
        field_norm = ((field_loc**2).sum())**0.5
        
        p_u = float(beam_polarization.polarization)
        p_d = (2.*float(beam_polarization.efficiency)-1.)*p_u
        
        e_u_loc = field_loc/field_norm

        refln = crystal.calc_sf(h, k, l)
        f_nucl, sft_11, sft_12, sft_13 = refln.f_nucl, refln.sft_11, refln.sft_12, refln.sft_13
        sft_21, sft_22, sft_23 = refln.sft_21, refln.sft_22, refln.sft_23
        sft_31, sft_32, sft_33 = refln.sft_31, refln.sft_32, refln.sft_33
        
        cell = crystal.cell
        k_1, k_2, k_3 = cell.calc_k_loc(h, k, l)
        
        mag_1 = sft_11*field_loc[0] + sft_12*field_loc[1] + sft_13*field_loc[2]
        mag_2 = sft_21*field_loc[0] + sft_22*field_loc[1] + sft_23*field_loc[2]
        mag_3 = sft_31*field_loc[0] + sft_32*field_loc[1] + sft_33*field_loc[2]
        
        #vector product k x mag x k        
        mag_p_1 = (k_3*mag_1 - k_1*mag_3)*k_3 - (k_1*mag_2 - k_2*mag_1)*k_2
        mag_p_2 = (k_1*mag_2 - k_2*mag_1)*k_1 - (k_2*mag_3 - k_3*mag_2)*k_3
        mag_p_3 = (k_2*mag_3 - k_3*mag_2)*k_2 - (k_3*mag_1 - k_1*mag_3)*k_1
        
        mag_p_sq = abs(mag_p_1*mag_p_1.conjugate() + mag_p_2*mag_p_2.conjugate() + 
                       mag_p_3*mag_p_3.conjugate())
        
        mag_p_e_u = mag_p_1*e_u_loc[0]+mag_p_2*e_u_loc[1]+mag_p_3*e_u_loc[2]
        
        f_nucl_sq = abs(f_nucl)**2
        mag_p_e_u_sq = abs(mag_p_e_u*mag_p_e_u.conjugate())
        fnp = (mag_p_e_u*f_nucl.conjugate()+mag_p_e_u.conjugate()*f_nucl).real
        fp_sq = f_nucl_sq + mag_p_sq + fnp
        fm_sq = f_nucl_sq + mag_p_sq - fnp
        fpm_sq = mag_p_sq - mag_p_e_u_sq

        #extinction correction     
        yp = extinction.calc_extinction(cell, h, k, l, fp_sq, wavelength)
        ym = extinction.calc_extinction(cell, h, k, l, fm_sq, wavelength)
        ypm = extinction.calc_extinction(cell, h, k, l, fpm_sq, wavelength)

        pppl = 0.5*((1+p_u)*yp+(1-p_u)*ym)
        ppmin= 0.5*((1-p_d)*yp+(1+p_d)*ym)
        pmpl = 0.5*((1+p_u)*yp-(1-p_u)*ym)
        pmmin= 0.5*((1-p_d)*yp-(1+p_d)*ym)
        """
        print("   h   k   l  f_nucl f_m_p_sq   f_np  fpm_sq")
        for h1, k1, l1, f_n_sq, f_m_sq, f_np, f_pm_sq in zip(h, k, l, f_nucl_sq, mag_p_e_u_sq, fnp, fpm_sq):
            print(" {:3} {:3} {:3} {:7.3f} {:7.3f} {:7.3f} {:7.3f}".format(
                    h1, k1, l1, f_n_sq, f_m_sq, f_np, f_pm_sq))

        print("   h   k   l      yp      ym     ypm      pppl   ppmin    pmpl   pmmin")
        for h1, k1, l1, y_p, y_m, y_pm, p_ppl, p_pmin, p_mpl, p_mmin in zip(
                h, k, l, yp, ym, ypm, pppl, ppmin, pmpl, pmmin):
            print(" {:3} {:3} {:3} {:7.3f} {:7.3f} {:7.3f}   {:7.3f} {:7.3f} {:7.3f} {:7.3f}".format(
                    h1, k1, l1, y_p, y_m, y_pm, p_ppl, p_pmin, p_mpl, p_mmin))
        """
        #integral intensities and flipping ratios
        iint_u = (f_nucl_sq+mag_p_e_u_sq)*pppl + pmpl*fnp + ypm*fpm_sq
        iint_d = (f_nucl_sq+mag_p_e_u_sq)*ppmin + pmmin*fnp + ypm*fpm_sq

        flip_ratio = iint_u/iint_d
        
        """
        d_info_out = {"iint_u": iint_u, "iint_d": iint_d, 
                      "flip_ratio": flip_ratio}
        d_info_out.update(d_info_sf)      
        print("   h   k   l  iint_u  iint_d flip_ratio")
        for h1, k1, l1, i_u, i_d, f_r in zip(h, k, l, iint_u, iint_d, flip_ratio):
            print("{:3} {:3} {:3} {:7.3f} {:7.3f} {:7.3f}".format(
                    h1, k1, l1, i_u, i_d, f_r))
        """
            
        return iint_u, iint_d, flip_ratio, refln
    
    def calc_chi_sq(self, l_crystal):
        """
        calculate chi square
        """
        diffrn_refln = self.diffrn_refln
        h, k, l = diffrn_refln.h, diffrn_refln.k, diffrn_refln.l
        fr_exp = diffrn_refln.fr
        fr_sigma = diffrn_refln.fr_sigma
        int_u_mod, int_d_mod, fr_mod, refln = self.calc_iint_u_d_flip_ratio(
                                               h, k, l, l_crystal)
        self.diffrn_refln.fr_calc = fr_mod
        self.diffrn_refln.intensity_up_calc = int_u_mod
        self.diffrn_refln.intensity_down_calc = int_d_mod

        self.refln = refln

        chi_sq = ((fr_mod-fr_exp)/fr_sigma)**2
        chi_sq_val = (chi_sq[numpy.logical_not(numpy.isnan(chi_sq))]).sum()
        n = numpy.logical_not(numpy.isnan(chi_sq)).sum()
        return chi_sq_val, n
    
    @property
    def is_variable(self):
        """
        without extinction
        """
        l_bool = []
        if self.beam_polarization is not None:
            l_bool.append(self.beam_polarization.is_variable)
        if self.extinction is not None:
            l_bool.append(self.extinction.is_variable)
        res = any(l_bool)
        return res

    def get_variables(self):
        l_variable = []
        if self.beam_polarization is not None:
            l_variable.extend(self.beam_polarization.get_variables())
        if self.extinction is not None:
            l_variable.extend(self.extinction.get_variables())
        return l_variable


    @property
    def to_cif(self):
        ls_out = []
        ls_out.append("data_{:}".format(self.label))
        str_1 = self.params_to_cif
        if str_1 != "":
            ls_out.append(str_1)
        str_2 = self.data_to_cif
        if str_2 != "":
            ls_out.append(str_2)
        str_3 = self.calc_to_cif
        if str_3 != "":
            ls_out.append(str_3)
        return "\n".join(ls_out)

    @property
    def params_to_cif(self):
        ls_out = []
        if self.extinction is not None:
            ls_out.append("\n"+self.extinction.to_cif)
        if self.beam_polarization is not None:
            ls_out.append("\n"+self.beam_polarization.to_cif)
        if self.phase_label is not None:
            ls_out.append("\n"+"_diffrn_phase_label {:}".format(self.phase_label))
        if self.wavelength is not None:
            ls_out.append("\n"+"_diffrn_radiation_wavelength {:}".format(self.wavelength))
        if self.field is not None:
            ls_out.append("_diffrn_ambient_field {:}".format(self.field))
        if self.orient_matrix is not None:
            ls_out.append("\n"+self.orient_matrix.to_cif)
        return "\n".join(ls_out)

    @property
    def data_to_cif(self):
        ls_out = []
        if self.diffrn_refln is not None:
            ls_out.append("\n"+self.diffrn_refln.to_cif)
        return "\n".join(ls_out)

    @property
    def calc_to_cif(self):
        ls_out = []
        #it is doubled with data_to_cif as there is calculated values.  Should it be separated???
        if self.diffrn_refln is not None:
            ls_out.append("\n"+self.diffrn_refln.to_cif)
        if self.refln is not None:
            ls_out.append("\n"+self.refln.to_cif)
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_data = Data()
        flag = cif_data.take_from_string(string)
        if not flag:
            return False
        self.label = cif_data.name
        cif_values = cif_data.items
        if cif_values is not None:
            if cif_values.is_prefix("_diffrn_phase_label"):
                self.phase_label = cif_values["_diffrn_phase_label"]
            if cif_values.is_prefix("_diffrn_radiation_wavelength"):
                self.wavelength = float(cif_values["_diffrn_radiation_wavelength"])
            if cif_values.is_prefix("_diffrn_ambient_field"):
                self.field = float(cif_values["_diffrn_ambient_field"])
            if cif_values.is_prefix("_diffrn_radiation"):
                self.beam_polarization = str(cif_values)
            if cif_values.is_prefix("_diffrn_orient_matrix_UB"):
                self.orient_matrix = str(cif_values)
            if cif_values.is_prefix("_refine_ls_extinction"):
                self.extinction = str(cif_values)
        if cif_data.is_prefix("_diffrn_refln"): self.diffrn_refln = str(cif_data["_diffrn_refln"])
        return True




"""
define classes to describe _orient_matrix_UB and  _orient_matrix_U
"""
__author__ = 'ikibalin'
__version__ = "2019_09_02"
import os
import numpy
import copy

from pycifstar import Global
from cryspy.f_common.cl_fitable import Fitable
from cryspy.f_crystal.cl_cell import Cell


class OrientMatrix(object):
    """
    describe orientation matrix
    """
    def __init__(self,  u=numpy.array([[1., 0., 0.],[0., 1., 0.],[0., 0., 1.]], dtype=float),
                        cell=Cell(), ub_from=None, ub_ccsl=None, gamma_0=None, nu_0=None, wavelength=None):
        super(OrientMatrix, self).__init__()
        self.__orient_matrix_ub_11 = None
        self.__orient_matrix_ub_12 = None
        self.__orient_matrix_ub_13 = None
        self.__orient_matrix_ub_21 = None
        self.__orient_matrix_ub_22 = None
        self.__orient_matrix_ub_23 = None
        self.__orient_matrix_ub_31 = None
        self.__orient_matrix_ub_32 = None
        self.__orient_matrix_ub_33 = None

        self.__orient_matrix_u_11 = None
        self.__orient_matrix_u_12 = None
        self.__orient_matrix_u_13 = None
        self.__orient_matrix_u_21 = None
        self.__orient_matrix_u_22 = None
        self.__orient_matrix_u_23 = None
        self.__orient_matrix_u_31 = None
        self.__orient_matrix_u_32 = None
        self.__orient_matrix_u_33 = None

        self.__gamma_0 = None
        self.__nu_0 = None
        self.__wavelength = None
        
        flag_ub_from = ub_from is not None
        flag_ub_ccsl = ub_ccsl is not None

        if flag_ub_from:
            self.ub_from = ub_from
        elif flag_ub_ccsl:
            self.ub_ccsl = ub_ccsl
        else:
            self.cell = cell
            self.u = u
        self.gamma_0 = gamma_0
        self.nu_0 = nu_0
        self.wavelength = wavelength

    @property
    def gamma_0(self):
        return self.__gamma_0
    @gamma_0.setter
    def gamma_0(self, x):
        if x is None:
            x_in = 0.
        else:
            x_in = float(x)
        self.__gamma_0 = x_in

    @property
    def nu_0(self):
        return self.__nu_0
    @nu_0.setter
    def nu_0(self, x):
        if x is None:
            x_in = 0.
        else:
            x_in = float(x)
        self.__nu_0 = x_in

    @property
    def wavelength(self):
        return self.__wavelength
    @wavelength.setter
    def wavelength(self, x):
        if x is None:
            x_in = 1.4
        else:
            x_in = float(x)
        self.__wavelength = x_in



    @property
    def u_11(self):
        return self.__orient_matrix_u_11
    @property
    def u_12(self):
        return self.__orient_matrix_u_12
    @property
    def u_13(self):
        return self.__orient_matrix_u_13

    @property
    def u_21(self):
        return self.__orient_matrix_u_21
    @property
    def u_22(self):
        return self.__orient_matrix_u_22
    @property
    def u_23(self):
        return self.__orient_matrix_u_23

    @property
    def u_31(self):
        return self.__orient_matrix_u_31
    @property
    def u_32(self):
        return self.__orient_matrix_u_32
    @property
    def u_33(self):
        return self.__orient_matrix_u_33

    @property
    def ub_11(self):
        return self.__orient_matrix_ub_11
    @property
    def ub_12(self):
        return self.__orient_matrix_ub_12
    @property
    def ub_13(self):
        return self.__orient_matrix_ub_13

    @property
    def ub_21(self):
        return self.__orient_matrix_ub_21
    @property
    def ub_22(self):
        return self.__orient_matrix_ub_22
    @property
    def ub_23(self):
        return self.__orient_matrix_ub_23

    @property
    def ub_31(self):
        return self.__orient_matrix_ub_31
    @property
    def ub_32(self):
        return self.__orient_matrix_ub_32
    @property
    def ub_33(self):
        return self.__orient_matrix_ub_33


    @property
    def ub_ccsl(self):
        """
        The UB matrix (CCSL)
        """
        ub = numpy.array([[self.ub_11, self.ub_12, self.ub_13], 
                          [self.ub_21, self.ub_22, self.ub_23], 
                          [self.ub_31, self.ub_32, self.ub_33]], dtype=float)
        return ub
    @ub_ccsl.setter
    def ub_ccsl(self, x):
        if isinstance(x, numpy.ndarray):
            x_in = x
        else:
            try:
                x_in = numpy.array(x, dtype=float)
            except:
                self._show_message("The UB matrix should be numpy.ndarray with size(3x3)")
                return
        flag = x_in.shape == (3,3)
        if flag:
            self.__orient_matrix_ub_11 = float(x_in[0,0])
            self.__orient_matrix_ub_12 = float(x_in[0,1])
            self.__orient_matrix_ub_13 = float(x_in[0,2])
            self.__orient_matrix_ub_21 = float(x_in[1,0])
            self.__orient_matrix_ub_22 = float(x_in[1,1])
            self.__orient_matrix_ub_23 = float(x_in[1,2])
            self.__orient_matrix_ub_31 = float(x_in[2,0])
            self.__orient_matrix_ub_32 = float(x_in[2,1])
            self.__orient_matrix_ub_33 = float(x_in[2,2])
            self.__recalc_u_cell
        else:
                self._show_message("The UB matrix should be numpy.ndarray with size(3x3)")
                return

    @property
    def cell(self):
        """
        The cell
        """
        return self.__cell
    @cell.setter
    def cell(self, x):
        if isinstance(x, Cell):
            x_in = x
        else:
            self._show_message("A type of induced element is not recognized to convert it into Cell")
            x_in=Cell()
        self.__cell = x_in

    @property
    def ub_from(self):
        """
        The UB matrix (from) 
        """
        #not sure it should be checked
        ub = numpy.array([[-1.*self.ub_21, -1.*self.ub_22, -1.*self.ub_23], 
                          [self.ub_11, self.ub_12, self.ub_13], 
                          [self.ub_31, self.ub_32, self.ub_33]], dtype=float)
        return ub
    @ub_from.setter
    def ub_from(self, x):
        self.__from_ub_from_to_ub_ccsl(x)

    @property
    def u(self):
        """
        The U matrix (orientation matrix) 
        """
        #not sure it should be checked
        u = numpy.array([[self.u_11, self.u_12, self.u_13], 
                         [self.u_21, self.u_22, self.u_23], 
                         [self.u_31, self.u_32, self.u_33]], dtype=float)
        return u
    @u.setter
    def u(self, x):
        if isinstance(x, numpy.ndarray):
            x_in = x
        else:
            try:
                x_in = numpy.array(x, dtype=float)
            except:
                self._show_message("The U matrix should be numpy.ndarray with size(3x3)")
                return
        flag = x_in.shape == (3,3)
        if flag:
            self.__orient_matrix_u_11 = float(x_in[0,0])
            self.__orient_matrix_u_12 = float(x_in[0,1])
            self.__orient_matrix_u_13 = float(x_in[0,2])
            self.__orient_matrix_u_21 = float(x_in[1,0])
            self.__orient_matrix_u_22 = float(x_in[1,1])
            self.__orient_matrix_u_23 = float(x_in[1,2])
            self.__orient_matrix_u_31 = float(x_in[2,0])
            self.__orient_matrix_u_32 = float(x_in[2,1])
            self.__orient_matrix_u_33 = float(x_in[2,2])
            self.__recalc_ub
        else:
                self._show_message("The U matrix should be numpy.ndarray with size(3x3)")
                return



    def __from_ub_from_to_ub_ccsl(self, ub_from):
        ub_ccsl = numpy.array([[ub_from[1, 0], ub_from[1, 1], ub_from[1, 2]],
                               [-1.*ub_from[0, 0], -1.*ub_from[0, 1], -1.*ub_from[0, 2]],
                               [ub_from[2, 0], ub_from[2, 1], ub_from[2, 2]]], dtype=float)
        self.ub_ccsl = ub_ccsl


    @property
    def __recalc_u_cell(self):
        ub_from = self.ub_from
    
        v_b_1 = ub_from[:, 0]
        v_b_2 = ub_from[:, 1]
        v_b_3 = ub_from[:, 2]

        b_1 = float(numpy.sqrt(((v_b_1*v_b_1).sum())))
        b_2 = float(numpy.sqrt(((v_b_2*v_b_2).sum())))
        b_3 = float(numpy.sqrt(((v_b_3*v_b_3).sum())))

        c_i_g = (v_b_1*v_b_2).sum()/(b_1*b_2)
        c_i_b = (v_b_1*v_b_3).sum()/(b_1*b_3)
        c_i_a = (v_b_2*v_b_3).sum()/(b_2*b_3)

        i_a, i_b, i_g = numpy.arccos(c_i_a), numpy.arccos(c_i_b), numpy.arccos(c_i_g)

        s_i_a, s_i_b, s_i_g = numpy.sin(i_a), numpy.sin(i_b), numpy.sin(i_g)
    
        i_vol=b_1*b_2*b_3*(1-c_i_a**2-c_i_b**2-c_i_g**2+2.*c_i_a*c_i_b*c_i_g)**0.5

        c_a = (c_i_b*c_i_g-c_i_a)/(s_i_b*s_i_g)
        c_b = (c_i_a*c_i_g-c_i_b)/(s_i_a*s_i_g)
        c_g = (c_i_a*c_i_b-c_i_g)/(s_i_a*s_i_b)

        a_1 = b_2*b_3*s_i_a/i_vol
        a_2 = b_1*b_3*s_i_b/i_vol
        a_3 = b_1*b_2*s_i_g/i_vol
        alpha, beta, gamma = numpy.arccos(c_a), numpy.arccos(c_b), numpy.arccos(c_g)

        cell = self.cell
        cell.a, cell.b, cell.c = a_1, a_2, a_3
        cell.alpha, cell.beta, cell.gamma = float(alpha*180./numpy.pi), float(beta*180./numpy.pi), float(gamma*180./numpy.pi),

        m_ib = cell.m_ib
        u = numpy.dot(ub_from, m_ib)
        
        self.__orient_matrix_u_11 = u[0, 0]
        self.__orient_matrix_u_12 = u[0, 1]
        self.__orient_matrix_u_13 = u[0, 2]
        self.__orient_matrix_u_21 = u[1, 0]
        self.__orient_matrix_u_22 = u[1, 1]
        self.__orient_matrix_u_23 = u[1, 2]
        self.__orient_matrix_u_31 = u[2, 0]
        self.__orient_matrix_u_32 = u[2, 1]
        self.__orient_matrix_u_33 = u[2, 2]

    @property
    def __recalc_ub(self):
        u = self.u
        cell = self.cell
        m_b = cell.m_b
        ub_from = numpy.dot(u, m_b)
        ub_ccsl = numpy.array([[ub_from[1, 0], ub_from[1, 1], ub_from[1, 2]],
                       [-1.*ub_from[0, 0], -1.*ub_from[0, 1], -1.*ub_from[0, 2]],
                       [ub_from[2, 0], ub_from[2, 1], ub_from[2, 2]]], dtype=float)

        self.__orient_matrix_ub_11 = ub_ccsl[0, 0]
        self.__orient_matrix_ub_12 = ub_ccsl[0, 1]
        self.__orient_matrix_ub_13 = ub_ccsl[0, 2]
        self.__orient_matrix_ub_21 = ub_ccsl[1, 0]
        self.__orient_matrix_ub_22 = ub_ccsl[1, 1]
        self.__orient_matrix_ub_23 = ub_ccsl[1, 2]
        self.__orient_matrix_ub_31 = ub_ccsl[2, 0]
        self.__orient_matrix_ub_32 = ub_ccsl[2, 1]
        self.__orient_matrix_ub_33 = ub_ccsl[2, 2]


    def __repr__(self):
        u = self.u
        ub_ccsl = self.ub_ccsl
        ub_from = self.ub_from
        cell = self.cell
        wavelength = self.wavelength
        gamma_0 = self.gamma_0
        nu_0 = self.nu_0
        ls_out = ["OrientMatrix:"]
        if wavelength is not None:
            ls_out.append("wavelength:   {:.3f} Angstrem".format(wavelength))
        if gamma_0 is not None:
            ls_out.append("offset gamma: {:.3f} deg.".format(gamma_0))
        if nu_0 is not None:
            ls_out.append("offset nu:    {:.3f} deg.".format(nu_0))

        ls_out.extend(["\nOrientation matrix U:\n{:12.5f}{:12.5f}{:12.5f}\n{:12.5f}{:12.5f}{:12.5f}\n{:12.5f}{:12.5f}{:12.5f}".format(
                        u[0, 0], u[0, 1], u[0, 2], 
                        u[1, 0], u[1, 1], u[1, 2], 
                        u[2, 0], u[2, 1], u[2, 2])])
        ls_out.extend(["\nThe UB matrix (CCSL):\n{:12.5f}{:12.5f}{:12.5f}\n{:12.5f}{:12.5f}{:12.5f}\n{:12.5f}{:12.5f}{:12.5f}".format(
                        ub_ccsl[0, 0], ub_ccsl[0, 1], ub_ccsl[0, 2], 
                        ub_ccsl[1, 0], ub_ccsl[1, 1], ub_ccsl[1, 2], 
                        ub_ccsl[2, 0], ub_ccsl[2, 1], ub_ccsl[2, 2])])
        ls_out.extend(["\nThe UB matrix (from):\n{:12.5f}{:12.5f}{:12.5f}\n{:12.5f}{:12.5f}{:12.5f}\n{:12.5f}{:12.5f}{:12.5f}".format(
                        ub_from[0, 0], ub_from[0, 1], ub_from[0, 2], 
                        ub_from[1, 0], ub_from[1, 1], ub_from[1, 2], 
                        ub_from[2, 0], ub_from[2, 1], ub_from[2, 2])])
        ls_out.append(str(cell))
        return "\n".join(ls_out)

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    @property
    def to_cif(self):
        ls_out = []
        ls_out.append("_diffrn_orient_matrix_UB_11 {:}".format(self.ub_11))
        ls_out.append("_diffrn_orient_matrix_UB_12 {:}".format(self.ub_12))
        ls_out.append("_diffrn_orient_matrix_UB_13 {:}".format(self.ub_13))
        ls_out.append("_diffrn_orient_matrix_UB_21 {:}".format(self.ub_21))
        ls_out.append("_diffrn_orient_matrix_UB_22 {:}".format(self.ub_22))
        ls_out.append("_diffrn_orient_matrix_UB_23 {:}".format(self.ub_23))
        ls_out.append("_diffrn_orient_matrix_UB_31 {:}".format(self.ub_31))
        ls_out.append("_diffrn_orient_matrix_UB_32 {:}".format(self.ub_32))
        ls_out.append("_diffrn_orient_matrix_UB_33 {:}".format(self.ub_33))
        return "\n".join(ls_out)


    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        ub_ccsl = numpy.zeros((3, 3), dtype=float)
        if cif_global.is_value("_diffrn_orient_matrix_UB_11"):
            ub_ccsl[0, 0] = float(cif_global["_diffrn_orient_matrix_UB_11"]) 
        if cif_global.is_value("_diffrn_orient_matrix_UB_12"):
            ub_ccsl[0, 1] = float(cif_global["_diffrn_orient_matrix_UB_12"]) 
        if cif_global.is_value("_diffrn_orient_matrix_UB_13"):
            ub_ccsl[0, 2] = float(cif_global["_diffrn_orient_matrix_UB_13"])

        if cif_global.is_value("_diffrn_orient_matrix_UB_21"):
            ub_ccsl[1, 0] = float(cif_global["_diffrn_orient_matrix_UB_21"]) 
        if cif_global.is_value("_diffrn_orient_matrix_UB_22"):
            ub_ccsl[1, 1] = float(cif_global["_diffrn_orient_matrix_UB_22"]) 
        if cif_global.is_value("_diffrn_orient_matrix_UB_23"):
            ub_ccsl[1, 2] = float(cif_global["_diffrn_orient_matrix_UB_23"])

        if cif_global.is_value("_diffrn_orient_matrix_UB_31"):
            ub_ccsl[2, 0] = float(cif_global["_diffrn_orient_matrix_UB_31"]) 
        if cif_global.is_value("_diffrn_orient_matrix_UB_32"):
            ub_ccsl[2, 1] = float(cif_global["_diffrn_orient_matrix_UB_32"]) 
        if cif_global.is_value("_diffrn_orient_matrix_UB_33"):
            ub_ccsl[2, 2] = float(cif_global["_diffrn_orient_matrix_UB_33"])
        self.ub_ccsl = ub_ccsl
        return True


    def read_ubfrom(self, f_name="ubfrom.raf"):
        fid = open(f_name)
        l_cont = fid.readlines()
        fid.close()
        wavelength = float(l_cont[0])
        ub_from = numpy.array([[float(hh_2) for hh_2 in hh_1.strip().split()] for hh_1 in l_cont[1:4]], dtype=float)
        gamma_0, nu_0 = [float(hh) for hh in l_cont[4].strip().split()]
        self.__gamma_0 = gamma_0
        self.__nu_0 = nu_0
        self.__wavelength = wavelength
        
        self.ub_from = ub_from

    def calc_angle(self, h,k,l):
        ub_from = self.ub_from
        q_ub = numpy.dot(ub_from, [h, k, l])
        q = numpy.linalg.norm(q_ub)
        wavelength = self.wavelength
        gamma_0 = self.gamma_0
        nu_0 = self.nu_0

        q_final = [- q**2 * wavelength / 2.0, 0, q_ub[2]]
        q_y2 = q**2 - q_final[0]**2 - q_final[2]**2
        flag_hh_2 = True

        if q_y2<0:
            print("Angles are not found.")
            flag_hh_2 = False
        if flag_hh_2:
            q_final[1] = numpy.sqrt(q_y2)
            k_f = [q_final[0] + 1.0/wavelength, q_final[1], q_final[2]]
            nu = numpy.arcsin(k_f[2]*wavelength)/numpy.pi*180. - nu_0
        
            gamma = numpy.arctan2(k_f[1],k_f[0])/numpy.pi*180. - gamma_0
            phi = (numpy.arctan2(q_final[1], q_final[0]) -numpy.arctan2(q_ub[1], q_ub[0]))/numpy.pi*180
            [phi, gamma, nu] = [phi if phi > 0. else phi + 360., gamma if gamma > 0. else gamma + 360., nu]
            print("gamma is {:7.3f}   nu is {:7.3f}   phi is {:7.3f}".format(gamma, nu, phi))



"""
define class PdPhase which describes the phase contributions in 1d powder diffraction experiment
"""
__author__ = 'ikibalin'
__version__ = "2019_09_06"
import os
import numpy


from pycifstar import Global
from cryspy.f_common.cl_fitable import Fitable


class PdPhase(object):
    """
    Example:

    loop_
    _pd_phase_label
    _pd_phase_scale
    _pd_phase_igsize
     Fe3O4 0.02381 0.0
    """
    def __init__(self, label=[], scale=[], igsize=[]):
        super(PdPhase, self).__init__()
        self.__pd_phase_label = []
        self.__pd_phase_scale = []
        self.__pd_phase_igsize = []

        self.label = label
        self.scale = scale
        self.igsize = igsize

    @property
    def label(self):
        return self.__pd_phase_label
    @label.setter
    def label(self, l_x):
        l_x_in = []
        for x in l_x:
            x_in = str(x).strip()
            l_x_in.append(x_in)
        self.__pd_phase_label = l_x_in
        len_x = len(l_x_in)

        len_1 = len(self.scale)
        if len_1 > len_x:
            self.__pd_phase_scale = self.__pd_phase_scale[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=1., name="_pd_phase_scale") for hh in range(len_x-len_1)] #default
            self.__pd_phase_scale.extend(l_fitable)

        len_1 = len(self.igsize)
        if len_1 > len_x:
            self.__pd_phase_igsize = self.__pd_phase_igsize[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_pd_phase_igsize") for hh in range(len_x-len_1)] #default
            self.__pd_phase_igsize.extend(l_fitable)

    @property
    def scale(self):
        return tuple(self.__pd_phase_scale)
    @scale.setter
    def scale(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                x_in.name = "_pd_phase_scale"
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=1., name="_pd_phase_scale") for hh in range(len_1-len_x)])
        self.__pd_phase_scale = l_fitable

    @property
    def igsize(self):
        return tuple(self.__pd_phase_igsize)
    @igsize.setter
    def igsize(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                x_in.name = "_pd_phase_igsize"
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name="_pd_phase_igsize") for hh in range(len_1-len_x)])
        self.__pd_phase_igsize = l_fitable


            
    def __repr__(self):
        ls_out = ["PdPhase:"]
        ls_out.append(" label   scale        igsize      ")
        for _1, _2, _3 in zip(self.label, self.scale, self.igsize):
            ls_out.append(" {:7} {:12} {:12}".format(_1, _2.print_with_sigma, _3.print_with_sigma))
        return "\n".join(ls_out)

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_pd_phase_label")
            ls_out.append("_pd_phase_scale")
            ls_out.append("_pd_phase_igsize")
            for _1, _2, _3 in zip(self.label, self.scale, self.igsize):
                ls_out.append("{:} {:} {:}".format(_1, _2.print_with_sigma, _3.print_with_sigma))
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_pd_phase")
        if flag:
            cif_loop = cif_global["_pd_phase"]
            l_name = cif_loop.names
            if "_pd_phase_label" in l_name:
                self.label = [_1.strip() for _1 in cif_loop["_pd_phase_label"]]
            if "_pd_phase_scale" in l_name:
                self.scale = [_1.strip() for _1 in cif_loop["_pd_phase_scale"]]
            if "_pd_phase_igsize" in l_name:
                self.igsize = [_1.strip() for _1 in cif_loop["_pd_phase_igsize"]]
        else:
            self.label = []
        return True

    @property
    def is_defined(self):
        cond = all([self.label is not None, self.scale is not None, self.igsize is not None])
        return cond

    @property
    def is_variable(self):
        res = (any([_1.refinement for _1 in self.scale]) |
               any([_1.refinement for _1 in self.igsize]))
        return res
    
    def get_variables(self):
        l_variable = []
        l_variable.extend([_1 for _1 in self.scale if _1.refinement])
        l_variable.extend([_1 for _1 in self.igsize if _1.refinement])
        return l_variable

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)



"""
define classes to calculated data
"""
__author__ = 'ikibalin'
__version__ = "2019_09_06"
import os
import numpy


from pycifstar import Data
from cryspy.f_common.cl_fitable import Fitable


#Description of setup class

class BeamPolarization(object):
    """
    Describe the beam polarisation

    Example:
    
    _diffrn_radiation_efficiency 1.0
    _diffrn_radiation_polarization -0.87

    """
    def __init__(self, polarization = Fitable(1.0), efficiency = Fitable(1.0)):
        super(BeamPolarization, self).__init__()
        self.__polarization = None
        self.__efficiency = None

        self.polarization = polarization
        self.efficiency = efficiency
        
    @property
    def polarization(self):
        """
        The polarization of the incident beam. 

        The permitted range is -1.0 -> 1.0
        """
        return self.__polarization
    @polarization.setter
    def polarization(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__polarization = x_in

    @property
    def efficiency(self):
        """
        The efficiency of the efficiency. 

        The permitted range is -1.0 -> 1.0
        """
        return self.__efficiency
    @efficiency.setter
    def efficiency(self, x):
        if isinstance(x, Fitable):
            x_in = x
        else:
            x_in = Fitable()
            flag = x_in.take_it(x)
        self.__efficiency = x_in


    def __repr__(self):
        ls_out = []
        ls_out.append("BeamPolarization: ")
        ls_out.append(" polarization: {}".format(self.polarization.print_with_sigma))
        ls_out.append(" efficiency: {}".format(self.efficiency.print_with_sigma))
        return "\n".join(ls_out)


    @property
    def is_variable(self):
        """
        Output: True if there is any refined parameter
        """
        res = any([self.polarization.refinement, 
                   self.efficiency.refinement])
        return res        
    
    def get_variables(self):
        """
        Output: the list of the refined parameters
        """
        l_variable = []
        if self.polarization.refinement:
            l_variable.append(self.polarization)
        if self.efficiency.refinement:
            l_variable.append(self.efficiency)
        return l_variable

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)
        
    @property
    def to_cif(self):
        ls_out = ["_diffrn_radiation_polarization {:}".format(self.polarization.print_with_sigma)]
        ls_out.append("_diffrn_radiation_efficiency {:}".format(self.efficiency.print_with_sigma))
        return "\n".join(ls_out)


    def from_cif(self, string: str):
        cif_data = Data()
        flag = cif_data.take_from_string(string)
        if not flag:
            return False
        flag = False
        if cif_data.is_value("_diffrn_radiation_polarization"):
            self.polarization = cif_data["_diffrn_radiation_polarization"] # CIFvalue
        if cif_data.is_value("_diffrn_radiation_efficiency"):
            self.efficiency = cif_data["_diffrn_radiation_efficiency"] # CIFvalue
        return True





"""
define classe to describe experiment
"""
__author__ = 'ikibalin'
__version__ = "2019_06_07"
import os
import numpy
import sys


from cryspy.f_experiment.f_single_domain.cl_observed_data_single_domain import ObservedDataSingleDomain
from cryspy.f_experiment.f_single.cl_calculated_data_single import CalculatedDataSingle
from cryspy.f_experiment.f_single.cl_setup_single import SetupSingle

from cryspy.f_common.cl_variable import Variable




class ExperimentSingleDomain(dict):
    """
    Class to describe all information concerning to the experiment for monocrystal
    """
    def __init__(self, name=None, setup = SetupSingle(), 
                 list_calculated_data = [], 
                 observed_data = ObservedDataSingleDomain(), file_out=None, file_dir=".", scale_domain=None):
        super(ExperimentSingleDomain, self).__init__()
        self._p_name = None
        self._p_setup = None
        self._list_calculated_data = []
        self._p_observed_data = None
        self._p_file_out = None
        self._p_file_dir = None
        self._p_scale_domain = []
        self._refresh(name, setup, observed_data, file_out, file_dir, scale_domain)

    def __repr__(self):
        ls_out = """ExperimentSingleDomain:\n name: {:}\n file_out: {:}\n{:}\n{:}""".format(
                self._p_name, self._p_file_out, self._p_setup, self._p_observed_data)
        
        ls_calculated_data = []
        for calculated_data in self._list_calculated_data:
            ls_calculated_data.append("{:}".format(calculated_data))

        ls_out += "\n\n\nCalculatedData:\n\n"+"\n\n".join(ls_calculated_data)
        return ls_out

    def _refresh(self, name, setup, observed_data, file_out, file_dir, scale_domain):
        if name is not None:
            self._p_name = name
        if setup is not None:
            self._p_setup = setup
        if observed_data is not None:
            self._p_observed_data = observed_data
        if file_out is not None:
            self._p_file_out = file_out
        if file_dir is not None:
            self._p_file_dir = file_dir
        if scale_domain is not None:
            self._p_scale_domain = [hh for hh in scale_domain]
            
    def set_val(self, name=None, setup=None, observed_data=None, file_out=None, file_dir=None, scale_domain=None):
        self._refresh(name, setup, observed_data, file_out, file_dir, scale_domain)
        
    def get_val(self, label):
        lab = "_p_"+label
        
        if lab in self.__dict__.keys():
            val = self.__dict__[lab]
            if isinstance(val, type(None)):
                self.set_val()
                val = self.__dict__[lab]
        else:
            print("The value '{:}' is not found".format(lab))
            val = None
        return val

    def list_vals(self):
        """
        give a list of parameters with small descripition
        """
        lsout = """
Parameters:
name is the name of experiment (should be unique)
setup is to describe parameters of diffractometer 
observed_data is the experimental data
file_out is the file name of model data
        """
        print(lsout)
    def add_calculated_data(self, observed_data):
        self._list_calculated_data.append(observed_data)

    def del_calculated_data(self, ind):
        self._list_calculated_data.pop(ind)        

    def replace_calculated_data(self, ind, observed_data):
        self._list_calculated_data.pop(ind)
        self._list_calculated_data.insert(ind, observed_data)
    
    def calc_iint_u_d_flip_ratio(self, h, k, l, l_crystal):
        """
        calculate intensity for the given diffraction angle
        """
        setup = self._p_setup
        wave_length = setup.get_val("wave_length")
        beam_polarization = setup.get_val("beam_polarization")
        
        #calculations only for first crystal phase
        for calculated_data in self._list_calculated_data[0:1]:
            ind_cry = None
            observed_data_name = calculated_data.get_val("name")
            for i_crystal, crystal in enumerate(l_crystal):
                if crystal.get_val("name") == observed_data_name:
                    ind_cry = i_crystal
                    break
            if ind_cry is None:
                print("Crystal with name '{:}' is not found.".format(
                    observed_data_name))
                return
            crystal = l_crystal[ind_cry]
            
            iint_u, iint_d, flip_ratio, d_info_cd = calculated_data.calc_iint_u_d_flip_ratio(
                              h, k, l, beam_polarization, wave_length, crystal)

        d_info_out = {}
        d_info_out.update(d_info_cd)        
            
        return iint_u, iint_d, flip_ratio, d_info_out
    
    def calc_chi_sq(self, l_crystal, d_info_in={}):
        """
        calculate chi square
        """
        #if not(d_map["flag"]|(d_map["out"] is None)):
        #    chi_sq_val, n = d_map["out"]
        #    return chi_sq_val, n        
        observed_data = self._p_observed_data

        np_h = observed_data.get_val('h')
        np_k = observed_data.get_val('k')
        np_l = observed_data.get_val('l')
        flip_ratio_exp = observed_data.get_val('flip_ratio')
        sflip_ratio_exp = observed_data.get_val('sflip_ratio')

        wave_length = observed_data.get_val('wave_length')
        setup = self._p_setup
        setup.set_val(wave_length=wave_length)

        field = observed_data.get_val('field')
        np_orientation = observed_data.get_val('orientation')


        n_domain = np_h.shape[-1]
        l_d_info_cd = []
        for i_domain in range(n_domain):
            h, k, l = np_h[:, i_domain], np_k[:, i_domain], np_l[:, i_domain]
            orientation = np_orientation[:, :, i_domain]
            
            for calculated_data in self._list_calculated_data:
                calculated_data.set_val(field=field, orientation=orientation)
            
            
            iint_u_mod, iint_d_mod, flip_ratio_mod, d_info_cd = self.calc_iint_u_d_flip_ratio(
                                               h, k, l, l_crystal)
            l_d_info_cd.append(d_info_cd)
            if i_domain == 0:
                np_iint_u = iint_u_mod[:, numpy.newaxis]
                np_iint_d = iint_d_mod[:, numpy.newaxis]
                np_flip_ratio = flip_ratio_mod[:, numpy.newaxis]
            else:
                np_iint_u = numpy.concatenate((np_iint_u, iint_u_mod[:, numpy.newaxis]), axis=1)
                np_iint_d = numpy.concatenate((np_iint_d, iint_d_mod[:, numpy.newaxis]), axis=1)
                np_flip_ratio = numpy.concatenate((np_flip_ratio, flip_ratio_mod[:, numpy.newaxis]), axis=1)
        
        
        l_scale_domain = [1.*hh for hh in self._p_scale_domain]
        if len(l_scale_domain) < n_domain:
            l_scale_domain.extend([1. for hh in range(n_domain-len(l_scale_domain))])
        np_scale_domain = numpy.array(l_scale_domain, dtype=float)
        l_flip_ratio_mod = []
        for iint_u_domain, iint_d_domain in zip(np_iint_u, np_iint_d):
            flag = numpy.logical_not(numpy.isnan(iint_u_domain))
            iint_u = (np_scale_domain[flag]*iint_u_domain[flag]).sum()
            iint_d = (np_scale_domain[flag]*iint_d_domain[flag]).sum()
            if all(list(numpy.logical_not(flag))):
                l_flip_ratio_mod.append(1.)
            else:
                l_flip_ratio_mod.append(iint_u/iint_d)

        flip_ratio_mod = numpy.array(l_flip_ratio_mod)
        chi_sq = ((flip_ratio_mod-flip_ratio_exp)/sflip_ratio_exp)**2
        
        chi_sq_val = (chi_sq[numpy.logical_not(numpy.isnan(chi_sq))]).sum()
        
        n = numpy.logical_not(numpy.isnan(chi_sq)).sum()

        d_info_out = {"chi_sq_val": chi_sq_val, "n": n, "l_d_info_cd": l_d_info_cd, 
        "iint_u": iint_u, "iint_d": iint_d, "flip_ratio": flip_ratio_mod}
        return chi_sq_val, n, d_info_out
    
    def plot_map(self):
        b_variable = self.is_variable()
        d_map = {"flag": b_variable, "out":None}
        return d_map
    
    def is_variable_iint(self):
        lres = []
        for calculated_data in self._list_calculated_data:
            lres.append(calculated_data.is_variable())
        return lres

    def is_variable_setup(self):
        setup = self.get_val("setup")
        res = setup.is_variable()
        return res

    def is_variable(self):
        """
        without extinction
        """
        lres = self.is_variable_iint() 
        lres.append(self.is_variable_setup())
        res = any(lres) 
        return res 
    
    def get_variables(self):
        l_variable = []
        l_scale_domain = self._p_scale_domain
        for sale_domain in l_scale_domain:
            if isinstance(sale_domain, Variable):
                l_variable.append(sale_domain)
        setup = self.get_val("setup")
        l_var = setup.get_variables()
        l_variable.extend(l_var)
        for calculated_data in self._list_calculated_data:
            l_var = calculated_data.get_variables()
            l_variable.extend(l_var)
        return l_variable

    def save_exp_mod_data(self, l_crystal):
        d_info_in = {}
        d_info_out = self.calc_chi_sq(l_crystal, d_info_in)[2]
        l_d_info_cd = d_info_out["l_d_info_cd"]

        fr_mod = d_info_out["flip_ratio"]
        observed_data = self.get_val("observed_data")
        fr_exp = observed_data.get_val("flip_ratio")
        sfr_exp = observed_data.get_val("sflip_ratio")
        
        line_head = []
        l_line_cont = [[] for hh in fr_exp]
        for i_domain, d_info_cd in enumerate(l_d_info_cd):
            line_head.append("  h_{0:}  k_{0:}  l_{0:}  iint_u_{0:}  iint_d_{0:}      fr_{0:}".format(i_domain+1))
            h_d, k_d, l_d = d_info_cd["h"], d_info_cd["k"], d_info_cd["l"]
            iint_u_d, iint_d_d = d_info_cd["iint_u"], d_info_cd["iint_d"]
            flip_ratio_d =  d_info_cd["flip_ratio"]
            for h, k, l, iint_u, iint_d, flip_ratio, l_line in zip(h_d, k_d, l_d, iint_u_d, iint_d_d, flip_ratio_d, l_line_cont):
                if numpy.isnan(h):
                    line=" None None None      None      None      None"
                else:
                    line="{:5}{:5}{:5}{:10.3f}{:10.3f}{:10.5f}".format(int(h), int(k), int(l), iint_u, iint_d, flip_ratio)
                l_line.append(line)
        line_head.append("      fr_mod    fr_exp   sfr_exp")
        for flip_ratio_mod, flip_ratio_exp, sflip_ratio_exp, l_line in zip(fr_mod, fr_exp, sfr_exp, l_line_cont):
            if numpy.isnan(flip_ratio_mod):
                line = "        None{:10.5f}{:10.5f}".format(flip_ratio_exp, sflip_ratio_exp)
            else:
                line = "  {:10.5f}{:10.5f}{:10.5f}".format(flip_ratio_mod, flip_ratio_exp, sflip_ratio_exp)
            l_line.append(line)

        ls_out = ["".join(line_head)]
        for l_line in l_line_cont:
            ls_out.append("".join(l_line))

        
        file_out, file_dir = self._p_file_out, self._p_file_dir
        if ((file_out is None) | (file_dir is None)):
            print("File to save model data is not defined.")
            return

        f_name = os.path.join(file_dir, file_out)
        fid = open(f_name, "w")
        fid.write("\n".join(ls_out))
        fid.close()

    def print_report(self, l_crystal):
        s_out = "{:}".format(self)
        d_info_in = {}
        d_info_out = self.calc_chi_sq(l_crystal, d_info_in)[2]
        l_d_info_cd = d_info_out["l_d_info_cd"]

        ls_out = []
        for i_domain, d_info_cd in enumerate(l_d_info_cd):
            ls_out.append("\n\n  h_{0:}  k_{0:}  l_{0:} fn_real_{0:} fn_imag_{0:}  iint_u_{0:}  iint_d_{0:}     fr_{0:}".format(i_domain+1))
            h_d, k_d, l_d = d_info_cd["h"], d_info_cd["k"], d_info_cd["l"]
            f_nucl_d = d_info_cd["f_nucl"]
            iint_u_d, iint_d_d, flip_ratio_d =  d_info_cd["iint_u"], d_info_cd["iint_d"], d_info_cd["flip_ratio"]
            sft_11_d, sft_12_d, sft_13_d = d_info_cd["sft_11"], d_info_cd["sft_12"], d_info_cd["sft_13"]
            sft_21_d, sft_22_d, sft_23_d = d_info_cd["sft_21"], d_info_cd["sft_22"], d_info_cd["sft_23"]
            sft_31_d, sft_32_d, sft_33_d = d_info_cd["sft_31"], d_info_cd["sft_32"], d_info_cd["sft_33"]


            for h, k, l, f, i_u, i_d, f_r in zip(
                h_d, k_d, l_d, f_nucl_d, iint_u_d, iint_d_d, flip_ratio_d):
                    if not(numpy.isnan(h)):
                        ls_out.append(""" {:4} {:4} {:4}  {:8.3f}  {:8.3f}  {:8.1f}  {:8.1f} {:8.5f}""".format(
                                  int(h), int(k), int(l), f.real, f.imag, i_u, i_d, f_r))
            ls_out.append("\n\nStructure factor tensor in a*, b*, c* (real part, mu_B) ")
            ls_out.append("\n   h   k   l   sft_11   sft_22   sft_33    sft_12   sft_13   sft_23")
            for h, k, l, c11, c22,  c33, c12, c13, c23 in zip(
                h_d, k_d, l_d, sft_11_d, sft_22_d, sft_33_d, sft_12_d, sft_13_d, sft_23_d):
                    if not(numpy.isnan(h)):
                        ls_out.append(""" {:3} {:3} {:3} {:8.3f} {:8.3f} {:8.3f}  {:8.3f} {:8.3f} {:8.3f}""".format(
                            int(h), int(k), int(l), c11.real, c22.real, c33.real, c12.real, c13.real, c23.real))

        return s_out+"\n".join(ls_out)


    
if (__name__ == "__main__"):
  pass



import os
import os.path

import sys

from cryspy.scripts.rhochi.cl_rhochi import  create_temporary

def main(l_arg= []):
    f_name_in = os.path.join(os.getcwd(), "main.rcif")
    create_temporary(f_name_in=f_name_in, exp_type="1")

if __name__ == '__main__':
    l_arg = sys.argv
    main(l_arg)


"""
define classes to describe AtomSite
"""
__author__ = 'ikibalin'
__version__ = "2019_08_30"
import os
import numpy


from pycifstar import Global
from cryspy.f_common.cl_fitable import Fitable
from cryspy.f_crystal.cl_fract import Fract

class AtomSite(object):
    """
    Data items in the ATOM_SITE category record details about
    the atom sites in a crystal structure, such as the positional
    coordinates.

    
    Description in cif file:

    loop_                                     
    _atom_site_label          
    _atom_site_type_symbol   
    _atom_site_fract_x       
    _atom_site_fract_y       
    _atom_site_fract_z       
    _atom_site_adp_type       
    _atom_site_B_iso_or_equiv
    _atom_site_occupancy     
     Fe3A   Fe  0.12500 0.12500 0.12500  uani   0.0   1.0
     Fe3B   Fe  0.50000 0.50000 0.50000  uani   0.0   1.0
     O1     O   0.25521 0.25521 0.25521  uiso   0.0   1.0

    
    reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Catom_site.html
    """    
    def __init__(self, label=[],  type_symbol=[], x=[], y=[], z=[], 
                       adp_type=[], b_iso=[], occupancy=[],
                       f_dir_prog = os.path.dirname(__file__)):
        super(AtomSite, self).__init__()

        self.__atom_site_label = []
        self.__atom_site_type_symbol = []
        self.__atom_site_fract_x = []
        self.__atom_site_fract_y = []
        self.__atom_site_fract_z = []
        self.__atom_site_adp_type = []
        self.__atom_site_b_iso_or_equiv = []
        self.__atom_site_occupancy = []
        self.__f_dir_prog = None
        #internal data
        self.__handbook_nucl = None
        self.__handbook_mag = None
        self.__scat_length_neutron = None
        self.__j0_A = None
        self.__j0_a = None
        self.__j0_B = None
        self.__j0_b = None
        self.__j0_C = None
        self.__j0_c = None
        self.__j0_D = None
        self.__j2_A = None
        self.__j2_a = None
        self.__j2_B = None
        self.__j2_b = None
        self.__j2_C = None
        self.__j2_c = None
        self.__j2_D = None

        self.__atom_site_multiplicity = None #not sure that it is needed

        self.label = label
        self.type_symbol = type_symbol
        self.x = x
        self.y = y
        self.z = z
        self.adp_type = adp_type
        self.b_iso = b_iso
        self.occupancy = occupancy
        self.f_dir_prog = f_dir_prog
        
    def __repr__(self):
        ls_out = ["AtomSite:"]
        ls_out.append(" label type_symbol x        y        z        adp_type b_iso    occupancy")
        for hh_1, hh_2, hh_3, hh_4, hh_5, hh_6, hh_7, hh_8 in zip(
            self.label, self.type_symbol, self.x, self.y, self.z, self.adp_type, self.b_iso, self.occupancy):
            ls_out.append(" {:8} {:8} {:8} {:8} {:8} {:8} {:8} {:8}".format(hh_1, hh_2,
                    hh_3.print_with_sigma, hh_4.print_with_sigma, hh_5.print_with_sigma,
                    hh_6, hh_7.print_with_sigma, hh_8.print_with_sigma))
        l_type_symbol = self.type_symbol
        l_ind = [l_type_symbol.index(hh) for hh in set(l_type_symbol)]
        l_scat = self.scat_length_neutron
        l_j0_A, l_j0_a, l_j0_B, l_j0_b, l_j0_C, l_j0_c, l_j0_D, l_j2_A, l_j2_a, l_j2_B, l_j2_b, l_j2_C, l_j2_c, l_j2_D = self.j0j2
        l_1, l_2, l_3 = [l_type_symbol[hh] for hh in l_ind], [l_scat[hh] for hh in l_ind], [l_j0_A[hh] for hh in l_ind]
        l_4, l_5, l_6 = [l_j0_a[hh] for hh in l_ind], [l_j0_B[hh] for hh in l_ind], [l_j0_b[hh] for hh in l_ind]
        l_7, l_8, l_9 = [l_j0_C[hh] for hh in l_ind], [l_j0_c[hh] for hh in l_ind], [l_j0_D[hh] for hh in l_ind]
        l_10, l_11, l_12 = [l_j2_A[hh] for hh in l_ind], [l_j2_a[hh] for hh in l_ind], [l_j2_B[hh] for hh in l_ind]
        l_13, l_14, l_15 = [l_j2_b[hh] for hh in l_ind], [l_j2_C[hh] for hh in l_ind], [l_j2_c[hh] for hh in l_ind]
        l_16 = [l_j2_D[hh] for hh in l_ind]

        for hh_1, hh_2, hh_3, hh_4, hh_5, hh_6, hh_7, hh_8, hh_9, hh_10, hh_11, hh_12, hh_13, hh_14, hh_15, hh_16 in zip(
            l_1, l_2, l_3, l_4, l_5, l_6, l_7, l_8, l_9, l_10, l_11, l_12, l_13, l_14, l_15, l_16):
            ls_out.append("\n for type_symbol '{:}'".format(hh_1))
            ls_out.append(" scat_length_neutron: {:.4f} cm**-12".format(hh_2))
            ls_out.append(" j0_A:{:8.4f},  j0_a:{:8.4f},  j0_B:{:8.4f},  j0_b:{:8.4f}".format(hh_3, hh_4, hh_5, hh_6))
            ls_out.append(" j0_C:{:8.4f},  j0_c:{:8.4f},  j0_D:{:8.4f}".format(hh_7, hh_8, hh_9))
            ls_out.append(" j2_A:{:8.4f},  j2_a:{:8.4f},  j2_B:{:8.4f},  j2_b:{:8.4f}".format(hh_10, hh_11, hh_12, hh_13))
            ls_out.append(" j2_C:{:8.4f},  j2_c:{:8.4f},  j2_D:{:8.4f}".format(hh_14, hh_15, hh_16))
        
        return "\n".join(ls_out)

    @property
    def label(self):
        """
        The _atom_site_label is a unique identifier for a particular site
        in the crystal. 

        Type: char

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_label.html
        """
        return tuple(self.__atom_site_label)
    @label.setter
    def label(self, l_x):
        l_x_in = []
        for x in l_x:
            x_in = str(x).strip()
            l_x_in.append(x_in)
        self.__atom_site_label = l_x_in
        len_x = len(l_x_in)

        len_1 = len(self.__atom_site_type_symbol)
        if len_1 > len_x:
            self.__atom_site_type_symbol = self.__atom_site_type_symbol[:len_x]
        elif len_1 < len_x:
            l_fitable = ["Fe3+" for hh in range(len_x-len_1)] #default
            self.__atom_site_type_symbol.extend(l_fitable)

        len_1 = len(self.__atom_site_fract_x)
        if len_1 > len_x:
            self.__atom_site_fract_x = self.__atom_site_fract_x[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_fract_x") for hh in range(len_x-len_1)]
            self.__atom_site_fract_x.extend(l_fitable)

        len_1 = len(self.__atom_site_fract_y)
        if len_1 > len_x:
            self.__atom_site_fract_y = self.__atom_site_fract_y[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_fract_y") for hh in range(len_x-len_1)]
            self.__atom_site_fract_y.extend(l_fitable)

        len_1 = len(self.__atom_site_adp_type)
        if len_1 > len_x:
            self.__atom_site_adp_type = self.__atom_site_adp_type[:len_x]
        elif len_1 < len_x:
            l_fitable = ["uiso" for hh in range(len_x-len_1)] #deafault value
            self.__atom_site_adp_type.extend(l_fitable)

        len_1 = len(self.__atom_site_b_iso_or_equiv)
        if len_1 > len_x:
            self.__atom_site_b_iso_or_equiv = self.__atom_site_b_iso_or_equiv[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=0., name="_atom_site_b_iso_or_equiv") for hh in range(len_x-len_1)]
            self.__atom_site_b_iso_or_equiv.extend(l_fitable)

        len_1 = len(self.__atom_site_occupancy)
        if len_1 > len_x:
            self.__atom_site_occupancy = self.__atom_site_occupancy[:len_x]
        elif len_1 < len_x:
            l_fitable = [Fitable(value=1., name="_atom_site_occupancy") for hh in range(len_x-len_1)]
            self.__atom_site_occupancy.extend(l_fitable)

    @property
    def type_symbol(self):
        """
        A code to identify the atom species (singular or plural)
        occupying this site.
        This code must match a corresponding _atom_type_symbol. 

        Type: char

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_type_symbol.html
        """
        return tuple(self.__atom_site_type_symbol)
    @type_symbol.setter
    def type_symbol(self, l_x):
        l_x_in = []
        for x in l_x:
            x_in = str(x).strip()
            l_x_in.append(x_in)
        len_x = len(l_x_in)
        len_1 = len(self.__atom_site_type_symbol)
        if len_1 < len_x:
            l_x_in = l_x_in[:len_1]
        elif len_1 > len_x:
            l_x_in.extend(["Fe3+" for hh in range(len_1-len_x)])
        self.__atom_site_type_symbol = l_x_in
        self.__scat_length_neutron = None
        self.__j0_A = None
        self.__j0_a = None
        self.__j0_B = None
        self.__j0_b = None
        self.__j0_C = None
        self.__j0_c = None
        self.__j0_D = None
        self.__j2_A = None
        self.__j2_a = None
        self.__j2_B = None
        self.__j2_b = None
        self.__j2_C = None
        self.__j2_c = None
        self.__j2_D = None

    @property
    def x(self):
        """
        Atom-site coordinates as fractions of the _cell_length_ values.

        Default: 0.

        Type: float

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_fract_.html
        """
        return tuple(self.__atom_site_fract_x)
    @x.setter
    def x(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name="_atom_site_fract_x") for hh in range(len_1-len_x)])
        self.__atom_site_fract_x = l_fitable


    @property
    def y(self):
        """
        see help for x parameter
        """
        return tuple(self.__atom_site_fract_y)
    @y.setter
    def y(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name="_atom_site_fract_y") for hh in range(len_1-len_x)])
        self.__atom_site_fract_y = l_fitable


    @property
    def z(self):
        """
        see help for x parameter
        """
        return tuple(self.__atom_site_fract_z)
    @z.setter
    def z(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name="_atom_site_fract_z") for hh in range(len_1-len_x)])
        self.__atom_site_fract_z = l_fitable

    @property
    def adp_type(self):
        """
        A standard code used to describe the type of atomic displacement
        parameters used for the site. 

        Type: [uani, uiso, uovl, umpe, bani, biso, bovl] #supported only [uani, uiso]

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_adp_type.html
        """
        return tuple(self.__atom_site_adp_type)
    @adp_type.setter
    def adp_type(self, l_x):
        l_list = ["uani", "uiso"]
        l_x_in = []
        for x in l_x:
            x_in = str(x).strip().lower()
            if (x_in not in l_list): x_in = "uiso"
            l_x_in.append(x_in)
        len_x = len(l_x_in)
        len_1 = len(self.__atom_site_adp_type)
        if len_1 < len_x:
            l_x_in = l_x_in[:len_1]
        elif len_1 > len_x:
            l_x_in.extend(["uiso" for hh in range(len_1-len_x)])
        self.__atom_site_adp_type = l_x_in


    @property
    def b_iso(self):
        """
        Isotropic atomic displacement parameter, or equivalent isotropic
        atomic displacement parameter, B(equiv), in angstroms squared,
        calculated from anisotropic displacement components.

        B(equiv) = (1/3) sum~i~[sum~j~(B^ij^ a*~i~ a*~j~ a~i~ a~j~)]

        a     = the real-space cell lengths
        a*    = the reciprocal-space cell lengths
        B^ij^ = 8 pi^2^ U^ij^

        Ref: Fischer, R. X. & Tillmanns, E. (1988). Acta Cryst. C44,
             775-776.

        The IUCr Commission on Nomenclature recommends against the use
        of B for reporting atomic displacement parameters. U, being
        directly proportional to B, is preferred.

        Default: 0.

        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_B_iso_or_equiv.html
        """
        return tuple(self.__atom_site_b_iso_or_equiv)
    @b_iso.setter
    def b_iso(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=0., name="_atom_site_b_iso_or_equiv") for hh in range(len_1-len_x)])
        self.__atom_site_b_iso_or_equiv = l_fitable


    @property
    def occupancy(self):
        """
        The fraction of the atom type present at this site.
        The sum of the occupancies of all the atom types at this site
        may not significantly exceed 1.0 unless it is a dummy site. The
        value must lie in the 99.97% Gaussian confidence interval
        -3u =< x =< 1 + 3u. The _enumeration_range of 0.0:1.0 is thus
        correctly interpreted as meaning (0.0 - 3u) =< x =< (1.0 + 3u).

        Default: 1.0
        
        reference: https://www.iucr.org/__data/iucr/cifdic_html/1/cif_core.dic/Iatom_site_occupancy.html
        """
        return tuple(self.__atom_site_occupancy)
    @occupancy.setter
    def occupancy(self, l_x):
        l_fitable = []
        for x in l_x:
            if isinstance(x, Fitable):
                x_in = x
            else:
                x_in = Fitable()
                flag = x_in.take_it(x)
            if x_in.value < 0.: x_in.value = 0.
            if x_in.value > 0.: x_in.value = 1.
            l_fitable.append(x_in)
        len_x = len(l_fitable)
        len_1 = len(self.__atom_site_label)
        if len_1 < len_x:
            l_fitable = l_fitable[:len_1]
        elif len_1 > len_x:
            l_fitable.extend([Fitable(value=1., name="_atom_site_occupancy") for hh in range(len_1-len_x)])
        self.__atom_site_occupancy = l_fitable



    @property
    def multiplicity(self):
        """
        Atom-site multiplicity 
        """
        return self.__atom_site_multiplicity
    @multiplicity.setter
    def multiplicity(self, l_x):
        l_val = []
        if l_x is None:
            l_val = None
        else:
            l_val = [int(x) for x in l_x]
        self.__atom_site_multiplicity = l_val

    @property
    def f_dir_prog(self):
        """

        reference:
        """
        return self.__f_dir_prog
    @f_dir_prog.setter
    def f_dir_prog(self, x):
        self.__f_dir_prog = x
        self._load_handbook_n()    
        self._load_handbook_m()

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    @property
    def is_defined(self):
        """
        Output: True if all started parameters are given
        """
        cond = (self.label != [])
        return cond
    @property
    def is_variable(self):
        res = (any([hh.refinement for hh in self.x]) | 
               any([hh.refinement for hh in self.y]) |
               any([hh.refinement for hh in self.z]) |
               any([hh.refinement for hh in self.b_iso]) |
               any([hh.refinement for hh in self.occupancy]))
        return res


    def get_variables(self):
        l_variable = []
        l_variable.extend([hh for hh in self.x if hh.refinement])
        l_variable.extend([hh for hh in self.y if hh.refinement])
        l_variable.extend([hh for hh in self.z if hh.refinement])
        l_variable.extend([hh for hh in self.b_iso if hh.refinement])
        l_variable.extend([hh for hh in self.occupancy if hh.refinement])
        return l_variable

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_atom_site_label")
            ls_out.append("_atom_site_type_symbol")
            ls_out.append("_atom_site_fract_x")
            ls_out.append("_atom_site_fract_y")
            ls_out.append("_atom_site_fract_z")
            ls_out.append("_atom_site_adp_type")
            ls_out.append("_atom_site_B_iso_or_equiv")
            ls_out.append("_atom_site_occupancy")
            flag_multiplicity = False
            if self.multiplicity is not None:
                flag_multiplicity = True
                ls_out.append("_atom_site_multiplicity")
            if flag_multiplicity:
                for _1, _2, _3, _4, _5, _6, _7, _8, _9 in zip(self.label, self.type_symbol, 
                    self.x, self.y, self.z, self.adp_type, self.b_iso, self.occupancy, self.multiplicity):
                    ls_out.append("{:} {:} {:} {:} {:} {:} {:} {:} {:}".format(_1, _2, 
                            _3.print_with_sigma, _4.print_with_sigma, _5.print_with_sigma,
                            _6, _7.print_with_sigma, _8.print_with_sigma, _9))
            else:
                for hh_1, hh_2, hh_3, hh_4, hh_5, hh_6, hh_7, hh_8 in zip(self.label, self.type_symbol, 
                    self.x, self.y, self.z, self.adp_type, self.b_iso, self.occupancy):
                    ls_out.append("{:} {:} {:} {:} {:} {:} {:} {:}".format(hh_1, hh_2, 
                            hh_3.print_with_sigma, hh_4.print_with_sigma, hh_5.print_with_sigma,
                            hh_6, hh_7.print_with_sigma, hh_8.print_with_sigma))
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_atom_site")
        if flag:
            cif_loop = cif_global["_atom_site"]
            l_name = cif_loop.names
            if "_atom_site_label" in l_name:
                self.label = cif_loop["_atom_site_label"]
            if "_atom_site_type_symbol" in l_name:
                self.type_symbol = cif_loop["_atom_site_type_symbol"]
            if "_atom_site_fract_x" in l_name:
                l_val = cif_loop["_atom_site_fract_x"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_fract_x")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.x = l_fitable
            if "_atom_site_fract_y" in l_name:
                l_val = cif_loop["_atom_site_fract_y"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_fract_y")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.y = l_fitable
            if "_atom_site_fract_z" in l_name:
                l_val = cif_loop["_atom_site_fract_z"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_fract_z")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.z = l_fitable
            if "_atom_site_adp_type" in l_name:
                self.adp_type = cif_loop["_atom_site_adp_type"]
            if "_atom_site_b_iso_or_equiv" in l_name:
                l_val = cif_loop["_atom_site_b_iso_or_equiv"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_b_iso_or_equiv")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.b_iso = l_fitable
            if "_atom_site_occupancy" in l_name:
                l_val = cif_loop["_atom_site_occupancy"]
                l_fitable = []
                for val in l_val:
                    fitable = Fitable(name="_atom_site_occupancy")
                    fitable.take_it(val)
                    l_fitable.append(fitable)
                self.occupancy = l_fitable
        else:
            self.label = []
        return True

    def _load_handbook_n(self):
        f_name = os.path.join(self.__f_dir_prog, "tables", "bscat.tab")
        fid = open(f_name, 'r')
        lcont = fid.readlines()
        fid.close()
        lcont = [line for line in lcont if not(line.startswith("#"))]
        ldcard = []
        for line in lcont:
            lhelp = line.strip().split()
            
            sline = lhelp[2].replace("i","j")
            sline = sline.split("(")[0]
            try:
                if sline.rfind("j") != -1:
                    b_scat = 0.1*complex(sline)
                else:
                    b_scat = 0.1*float(sline)
            except:
                b_scat = 0.
            dcard = {"type_n": lhelp[0], "b_scat": b_scat}
            ldcard.append(dcard)
        self.__handbook_nucl = ldcard

    def _load_handbook_m(self):
        f_name = os.path.join(self.__f_dir_prog, "tables", "formmag.tab")
        fid = open(f_name, 'r')
        lcont = fid.readlines()
        fid.close()
        lcont = [line for line in lcont if line.startswith("F")]
        ldcard = []
        for line in lcont:
            lhelp = line.strip().split()
            dcard = {"type_m": lhelp[1], "order": int(lhelp[2]),
                     "A": float(lhelp[3]),"a": float(lhelp[4]),
                     "B": float(lhelp[5]),"b": float(lhelp[6]),
                     "C": float(lhelp[7]),"c": float(lhelp[8]),
                     "D": float(lhelp[9])}
            ldcard.append(dcard)
        self.__handbook_mag = ldcard

    @property
    def scat_length_neutron(self):
        if self.__scat_length_neutron is None:
            l_res = [self._get_scat_length_neutron(type_symbol) for type_symbol in self.type_symbol]
            res = tuple(l_res)
            self.__scat_length_neutron = l_res 
        else:
            res = tuple(self.__scat_length_neutron)
        return res

    @property
    def b_scat(self):
        return self.scat_length_neutron

    def _get_scat_length_neutron(self, type_n):
        """
        Take scat_length_neutron
        """
        str_1 = type_n.strip().lower()
        str_1 = "".join([hh if hh.isalpha() else ' ' for hh in str_1 ]).split(" ")[0]
        l_dcard = self.__handbook_nucl 
        flag = False
        for dcard in l_dcard:
            if (dcard["type_n"].lower() == str_1):
                res = dcard["b_scat"]
                flag = True
            elif flag:
                break
        if not(flag):
            res = 0.
            self._show_message("Can not find b_scat for '{:}'.\n It is putted as 0.".format(type_n))
        return res 


    @property
    def j0j2(self):
        flag = any([self.__j0_A is None, self.__j0_a is None, self.__j0_B is None, self.__j0_b is None, 
                    self.__j0_C is None, self.__j0_c is None, self.__j0_D is None, 
                    self.__j2_A is None, self.__j2_a is None, self.__j2_B is None, self.__j2_b is None, 
                    self.__j2_C is None, self.__j2_c is None, self.__j2_D is None])

        if flag:
            l_j0_A, l_j0_a, l_j0_B, l_j0_b, l_j0_C, l_j0_c, l_j0_D = [], [], [], [], [], [], []
            l_j2_A, l_j2_a, l_j2_B, l_j2_b, l_j2_C, l_j2_c, l_j2_D = [], [], [], [], [], [], []
            for type_symbol in self.type_symbol:
                j0_A, j0_a, j0_B, j0_b, j0_C, j0_c, j0_D, j2_A, j2_a, j2_B, j2_b, j2_C, j2_c, j2_D = self._get_j0j2(type_symbol)
                l_j0_A.append(j0_A)
                l_j0_a.append(j0_a)
                l_j0_B.append(j0_B)
                l_j0_b.append(j0_b)
                l_j0_C.append(j0_C)
                l_j0_c.append(j0_c)
                l_j0_D.append(j0_D)
                l_j2_A.append(j2_A)
                l_j2_a.append(j2_a)
                l_j2_B.append(j2_B)
                l_j2_b.append(j2_b)
                l_j2_C.append(j2_C)
                l_j2_c.append(j2_c)
                l_j2_D.append(j2_D)
            self.__j0_A, self.__j0_a, self.__j0_B, self.__j0_b = tuple(l_j0_A), tuple(l_j0_a), tuple(l_j0_B), tuple(l_j0_b)
            self.__j0_C, self.__j0_c, self.__j0_D = tuple(l_j0_C), tuple(l_j0_c), tuple(l_j0_D)
            self.__j2_A, self.__j2_a, self.__j2_B, self.__j2_b = tuple(l_j2_A), tuple(l_j2_a), tuple(l_j2_B), tuple(l_j2_b)
            self.__j2_C, self.__j2_c, self.__j2_D = tuple(l_j2_C), tuple(l_j2_c), tuple(l_j2_D)
            l_res = tuple([l_j0_A, l_j0_a, l_j0_B, l_j0_b, l_j0_C, l_j0_c, l_j0_D, 
                           l_j2_A, l_j2_a, l_j2_B, l_j2_b, l_j2_C, l_j2_c, l_j2_D])
        else:
            j0_A, j0_a, j0_B, j0_b = self.__j0_A, self.__j0_a, self.__j0_B, self.__j0_b
            j0_C, j0_c, j0_D = self.__j0_C, self.__j0_c, self.__j0_D
            j2_A, j2_a, j2_B, j2_b = self.__j2_A, self.__j2_a, self.__j2_B, self.__j2_b
            j2_C, j2_c, j2_D = self.__j2_C, self.__j2_c, self.__j2_D
            l_res = tuple([j0_A, j0_a, j0_B, j0_b, j0_C, j0_c, j0_D, 
                           j2_A, j2_a, j2_B, j2_b, j2_C, j2_c, j2_D])
        return l_res

    @property
    def j0_A(self):
        if self.__j0_A is None:
            l_res = self.j0j2
            res = l_res[0]
        else:
            res = self.__j0_A
        return res
    @property
    def j0_a(self):
        if self.__j0_a is None:
            l_res = self.j0j2
            res = l_res[1]
        else:
            res = self.__j0_a
        return res
    @property
    def j0_B(self):
        if self.__j0_B is None:
            l_res = self.j0j2
            res = l_res[2]
        else:
            res = self.__j0_B
        return res
    @property
    def j0_b(self):
        if self.__j0_b is None:
            l_res = self.j0j2
            res = l_res[3]
        else:
            res = self.__j0_b
        return res
    @property
    def j0_C(self):
        if self.__j0_C is None:
            l_res = self.j0j2
            res = l_res[4]
        else:
            res = self.__j0_C
        return res
    @property
    def j0_c(self):
        if self.__j0_c is None:
            l_res = self.j0j2
            res = l_res[5]
        else:
            res = self.__j0_c
        return res
    @property
    def j0_D(self):
        if self.__j0_D is None:
            l_res = self.j0j2
            res = l_res[6]
        else:
            res = self.__j0_D
        return res
    @property
    def j2_A(self):
        if self.__j2_A is None:
            l_res = self.j0j2
            res = l_res[7]
        else:
            res = self.__j2_A
        return res
    @property
    def j2_a(self):
        if self.__j2_a is None:
            l_res = self.j0j2
            res = l_res[8]
        else:
            res = self.__j2_a
        return res
    @property
    def j2_B(self):
        if self.__j2_B is None:
            l_res = self.j0j2
            res = l_res[9]
        else:
            res = self.__j2_B
        return res
    @property
    def j2_b(self):
        if self.__j2_b is None:
            l_res = self.j0j2
            res = l_res[10]
        else:
            res = self.__j2_b
        return res
    @property
    def j2_C(self):
        if self.__j2_C is None:
            l_res = self.j0j2
            res = l_res[11]
        else:
            res = self.__j2_C
        return res
    @property
    def j2_c(self):
        if self.__j2_c is None:
            l_res = self.j0j2
            res = l_res[12]
        else:
            res = self.__j2_c
        return res
    @property
    def j2_D(self):
        if self.__j2_D is None:
            l_res = self.j0j2
            res = l_res[13]
        else:
            res = self.__j2_D
        return res


    def _get_j0j2(self, type_m):
        """
        Take coefficients for <j0> and <j2>
        """
        str_1 = type_m.strip().lower()
        str_1 = "".join([hh if hh.isalnum() else ' ' for hh in str_1 ]).split(" ")[0]

        ldcard = self.__handbook_mag 
        flag_0, flag_2 = False, False
        for dcard in ldcard:
            if ((dcard["type_m"].lower() == str_1)&(dcard["order"] == 0)):
                j0_A = dcard["A"]
                j0_a = dcard["a"]
                j0_B = dcard["B"]
                j0_b = dcard["b"]
                j0_C = dcard["C"]
                j0_c = dcard["c"]
                j0_D = dcard["D"]
                flag_0 = True
            elif ((dcard["type_m"].lower() == str_1)&(dcard["order"] == 2)):
                j2_A = dcard["A"]
                j2_a = dcard["a"]
                j2_B = dcard["B"]
                j2_b = dcard["b"]
                j2_C = dcard["C"]
                j2_c = dcard["c"]
                j2_D = dcard["D"]
                flag_2 = True
            elif (flag_0 & flag_2):
                break
        if not(flag_0):
            j0_A, j0_a, j0_B, j0_b, j0_C, j0_c, j0_D = 0., 0., 0., 0., 0., 0. ,0.
            self._show_message("Can not find coefficients <j0> for '{:}'.\n It is setted as 0.".format(type_m))
        if not(flag_2):
            j2_A, j2_a, j2_B, j2_b, j2_C, j2_c, j2_D = 0., 0., 0., 0., 0., 0. ,0.
            self._show_message("Can not find coefficients <j2> for '{:}.\n It is setted as 0.'".format(type_m))
        return j0_A, j0_a, j0_B, j0_b, j0_C, j0_c, j0_D, j2_A, j2_a, j2_B, j2_b, j2_C, j2_c, j2_D

    def _form_fract(self):
        fract = Fract(x=numpy.array(self.x, dtype=float), 
                      y=numpy.array(self.y, dtype=float), 
                      z=numpy.array(self.z, dtype=float))
        return fract

    def calc_constr_number(self, space_group):
        """
        according to table 1 in Peterse, Palm, Acta Cryst.(1966), 20, 147
        """
        l_numb = []
        for _1, _2, _3 in zip(self.x, self.y, self.z):
            x, y, z = float(_1), float(_2), float(_3)
            o_11, o_12, o_13, o_21, o_22, o_23, o_31, o_32, o_33, o_3, o_2, o_3 = space_group.calc_el_symm_for_xyz(x,y,z)

            b_11, b_22, b_33, b_12, b_13, b_23 = 107, 181, 41, 7, 19, 1

            i_11 = (o_11*o_11*b_11 + o_12*o_12*b_22 + o_13*o_13*b_33 + 
                    o_11*o_12*b_12 + o_11*o_13*b_13 + o_12*o_13*b_23 + 
                    o_12*o_11*b_12 + o_13*o_11*b_13 + o_13*o_12*b_23)

            i_22 = (o_21*o_21*b_11 + o_22*o_22*b_22 + o_23*o_23*b_33 + 
                    o_21*o_22*b_12 + o_21*o_23*b_13 + o_22*o_23*b_23 + 
                    o_22*o_21*b_12 + o_23*o_21*b_13 + o_23*o_22*b_23)

            i_33 = (o_31*o_31*b_11 + o_32*o_32*b_22 + o_33*o_33*b_33 + 
                    o_31*o_32*b_12 + o_31*o_33*b_13 + o_32*o_33*b_23 + 
                    o_32*o_31*b_12 + o_33*o_31*b_13 + o_33*o_32*b_23) 

            i_12 = (o_11*o_21*b_11 + o_12*o_22*b_22 + o_13*o_23*b_33 + 
                    o_11*o_22*b_12 + o_11*o_23*b_13 + o_12*o_23*b_23 + 
                    o_12*o_21*b_12 + o_13*o_21*b_13 + o_13*o_22*b_23) 

            i_13 = (o_11*o_31*b_11 + o_12*o_32*b_22 + o_13*o_33*b_33 + 
                    o_11*o_32*b_12 + o_11*o_33*b_13 + o_12*o_33*b_23 + 
                    o_12*o_31*b_12 + o_13*o_31*b_13 + o_13*o_32*b_23)

            i_23 = (o_21*o_31*b_11 + o_22*o_32*b_22 + o_23*o_33*b_33 + 
                    o_21*o_32*b_12 + o_21*o_33*b_13 + o_22*o_33*b_23 + 
                    o_22*o_31*b_12 + o_23*o_31*b_13 + o_23*o_32*b_23) 

            r_11, r_22, r_33, r_12, r_13, r_23 = i_11.sum(), i_22.sum(), i_33.sum(), i_12.sum(), i_13.sum(), i_23.sum()

            if r_13 == 0:
                if r_23 == 0:
                    if r_12 == 0:
                        if r_11 == r_22:
                            if r_22 == r_33:
                                numb = 17
                            else:
                                numb = 8
                        elif r_22 == r_33:
                            numb = 12
                        else:
                            numb = 4
                    elif r_11 == r_22:
                        if r_22 == 2*r_12:
                            numb = 16
                        else:
                            numb = 5
                    elif r_22 == 2*r_12:
                        numb = 14
                    else:
                        numb = 2
                elif r_22 == r_33:
                    numb = 9
                else:
                    numb = 3
            elif r_23 == 0:
                if r_22 == 2*r_12:
                    numb = 13
                else:
                    numb = 1
            elif r_23 == r_13:
                if r_22 == r_33:
                    numb = 18
                else:
                    numb = 6
            elif r_12 == r_13:
                numb = 10
            elif r_11 == 22:
                numb = 7
            elif r_22 == r_33:
                numb = 11
            elif r_22 == 2*r_12:
                numb = 15
            else:
                numb = 0 #no constraint
            l_numb.append(numb)
        return l_numb


"""
define classe PdExclude2Theta which describes the excluded regions in 1d powder diffraction experiment
"""
__author__ = 'ikibalin'
__version__ = "2019_09_06"
import os
import numpy


from pycifstar import Global


class PdExclude2Theta(object):
    """
    Example:

    loop_
    _pd_exclude_2theta_min
    _pd_exclude_2theta_max
     4.0  12.0 
     30.0 45.0 
     58.0 63.0 
    """
    def __init__(self, min=[], max=[]
                 ):
        super(PdExclude2Theta, self).__init__()
        self.__pd_exclude_2theta_min = None
        self.__pd_exclude_2theta_max = None

        self.min = min
        self.max = max

    @property
    def min(self):
        return self.__pd_exclude_2theta_min
    @min.setter
    def min(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_exclude_2theta_min = np_x_in


    @property
    def max(self):
        return self.__pd_exclude_2theta_max
    @max.setter
    def max(self, l_x):
        l_x_in = []
        for x in l_x:
            if isinstance(x, float):
                x_in = x
            else:
                x_in = float(x)
            l_x_in.append(x_in)
        np_x_in = numpy.array(l_x_in, dtype=float)
        self.__pd_exclude_2theta_max = np_x_in


    def __repr__(self):
        ls_out = ["PdExclude2Theta:"]
        ls_out.append("    min     max")
        for _1, _2 in zip(self.min, self.max):
            ls_out.append("{:7.2f} {:7.2f}".format(_1, _2))
        return "\n".join(ls_out)

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("loop_")
            ls_out.append("_pd_exclude_2theta_min")
            ls_out.append("_pd_exclude_2theta_max")
            for _1, _2 in zip(self.min, self.max):
                ls_out.append("{:} {:}".format(_1, _2))
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_pd_exclude_2theta")
        if flag:
            cif_loop = cif_global["_pd_exclude_2theta"]
            l_name = cif_loop.names
            if "_pd_exclude_2theta_min" in l_name:
                self.min = [float(_1) for _1 in cif_loop["_pd_exclude_2theta_min"]]
            if "_pd_exclude_2theta_max" in l_name:
                self.max = [float(_1) for _1 in cif_loop["_pd_exclude_2theta_max"]]
        else:
            self.min, self.max = [], []
        return True

    @property
    def is_defined(self):
        cond = all([self.min is not None, self.max is not None])
        return cond

    @property
    def is_variable(self):
        return False
    
    def get_variables(self):
        return []

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)



__version__ = '1.0.0'
__author__ = 'Iurii Kibalin'
__contact__ = 'iurii.kibalin@cea.fr'

from cryspy.f_crystal.cl_space_group import SpaceGroup
from cryspy.f_crystal.cl_cell import Cell
from cryspy.f_crystal.cl_atom_type import AtomType
from cryspy.f_crystal.cl_fract import Fract
from cryspy.f_crystal.cl_adp import ADP
from cryspy.f_crystal.cl_magnetism import (
    Magnetism,
    calc_mRmCmRT
    )
from cryspy.f_crystal.cl_atom_site import AtomSite
from cryspy.f_crystal.cl_crystal import Crystal


from cryspy.f_crystal.cl_atom_site_magnetism import AtomSiteMagnetism
from cryspy.f_crystal.cl_atom_site_magnetism_aniso import AtomSiteMagnetismAniso
from cryspy.f_crystal.cl_atom_site_aniso import AtomSiteAniso






"""
define the class Fitable.
"""
__author__ = 'ikibalin'
__version__ = "2019_08_22"

import numpy
from math import *
from pycifstar import Item

class Fitable(object):
    """
    Fitable class to describe variable parameter
    """
    def __init__(self, value=0., sigma=0., refinement=False, name=None, 
        constraint_flag=False, constraint=None, minimal=None, maximal=None):
        super(Fitable, self).__init__()
        self.__value = None
        self.__sigma = None
        self.__refinement = None
        self.__constraint_flag = None
        self.__name = None
        self.__constraint = None
        self.__minimal = None
        self.__maximal = None
        self.value = value
        self.sigma = sigma
        self.refinement = refinement
        self.constraint_flag = constraint_flag
        self.name = name
        self.constraint = constraint
        self.minimal = minimal
        self.maximal = maximal

    def take_it(self, object_):
        """
        Try to convert some objects to Fitable attributes
        """
        if isinstance(object_, Fitable):
            self.value = object_.value
            self.sigma = object_.sigma
            self.refinement = object_.refinement
            self.constraint_flag = object_.constraint_flag
            self.name = object_.name
            self.constraint = object_.constraint
            self.minimal = object_.minimal
            self.maximal = object_.maximal
        elif isinstance(object_, Item):
            self.name = object_.name
            self.constraint = object_.comment
            string = object_.value
            ind_1 = string.find("(")
            if  ind_1 != -1:
                self.refinement = True
            str_1 = string.split("(")[0]
            try:
                self.value = float(str_1)
            except:
                self._show_message("Can not convert value of Item to float")
                return False
        elif isinstance(object_, str):
            string = object_
            ind_1 = string.find("(")
            if ind_1 != -1:
                self.refinement = True
            str_1 = string.split("(")[0]
            try:
                self.value = float(str_1)
            except:
                self._show_message("Can not convert string to float")
                return False
        elif (isinstance(object_, float) | isinstance(object_, int)):
            self.value = float(object_)
        else:
            self._show_message("Unsupported format object for convertation to Fitable")
            return False
        return True

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    @property
    def value(self):
        return self.__value
    @value.setter
    def value(self, x):
        try:
            self.__value = float(x)
        except:
            self.__value = None
    @property
    def sigma(self):
        return self.__sigma
    @sigma.setter
    def sigma(self, x):
        try:
            self.__sigma = float(x)
        except:
            self.__sigma = None
    @property
    def refinement(self):
        return self.__refinement
    @refinement.setter
    def refinement(self, x):
        try:
            self.__refinement = bool(x)
        except:
            self.__refinement = False
    @property
    def constraint_flag(self):
        return self.__constraint_flag
    @constraint_flag.setter
    def constraint_flag(self, x):
        try:
            self.__constraint_flag = bool(x)
        except:
            self.__constraint_flag = False
    @property
    def name(self):
        return self.__name
    @name.setter
    def name(self, x):
        if x is not None:
            try:
                self.__name = str(x)
            except:
                self.__name = None
        else:
            self.__name = None
    @property
    def constraint(self):
        return self.__constraint
    @constraint.setter
    def constraint(self, x):
        if x is not None:
            try:
                self.__constraint = str(x)
            except:
                self.__constraint = None
        else:
            self.__constraint = None
    @property
    def minimal(self):
        return self.__minimal
    @minimal.setter
    def minimal(self, x):
        try:
            self.__minimal = float(x)
        except:
            self.__minimal = None
    @property
    def maximal(self):
        return self.__maximal
    @maximal.setter
    def maximal(self, x):
        try:
            self.__maximal = float(x)
        except:
            self.__maximal = None
    def __array__(self):
        return numpy.array(self.value)
    def __float__(self):
        return self.value
    def __bool__(self):
        return self.refinement
    def __repr__(self):
        ls_out = [self.print_with_sigma]
        if self.name is not None:
            ls_out.append("name: {:}".format(self.name))
        if self.minimal is not None:
            ls_out.append("minimal: {:.3f}".format(self.minimal))
        if self.maximal is not None:
            ls_out.append("maximal: {:.3f}".format(self.maximal))
        if self.refinement:
            ls_out.append("refinement: {:}".format(self.refinement))
        if (self.constraint is not None):
            if self.constraint != "":
                ls_out.append("constraint: {:}".format(self.constraint))
        res = ls_out[0]
        if len(ls_out) > 1:
            res = "{:} ({:})".format(res, ", ".join(ls_out[1:]))            
        return res
    @property
    def print_with_sigma(self):
        if not((self.sigma == 0.)|(self.sigma is None)):
            val_hh = numpy.log10(self.sigma)
            n_power = int(val_hh)
            if val_hh <= 0:
                val_1 = numpy.round(self.value, decimals = -1*int(n_power)+2)
                val_2 = numpy.round(self.sigma, decimals = -1*int(n_power)+2)
                i_val_11 = int(val_1)
                s_sign = ""
                if val_1 < 0.:
                    s_sign = "-"
                s_val_12 = ("{:}".format(int(abs(val_1)%1.*10**(-n_power+2)))).rjust(int(-n_power+2),"0")
                val_2=int(val_2*10**(-n_power+2))
                ls_out = " {:}{:}.{:}({:})".format(s_sign, abs(i_val_11), s_val_12, int(val_2))
            else:
                val_1 = numpy.round(self.value)
                val_2 = numpy.round(self.sigma)
                ls_out = " {:}({:})".format(int(val_1), int(val_2))
        elif self.refinement:
            ls_out = "{:}()".format(self.value)
        else:
            ls_out = "{:}".format(self.value)
        return ls_out 
    def __add__(self, var2):
        if type(var2) is numpy.ndarray:
            res_1 = self.value+var2
            res = res_1.astype(var2.dtype)
        elif isinstance(var2, Fitable):
            res = Fitable(value=self.value+var2.value,
                          refinement=(self.refinement|var2.refinement),
                          sigma=(self.sigma**2+var2.sigma**2)**0.5
                          )
        else:
            res = self.value+var2
        return res
    def __radd__(self, var2):
        if type(var2) is numpy.ndarray:
            res_1 = var2+self.value
            res = res_1.astype(var2.dtype)
        elif isinstance(var2, Fitable):
            res = Fitable(value=self.value+var2.value,
                          refinement=(self.refinement|var2.refinement),
                          sigma=(self.sigma**2+var2.sigma**2)**0.5
                          )
        else:
            res = var2+self.value
        return res
    def __sub__(self, var2):
        if type(var2) is numpy.ndarray:
            res_1 = self.value-var2
            res = res_1.astype(var2.dtype)
        elif isinstance(var2, Fitable):
            res = Fitable(value=self.value-var2.value,
                          refinement=(self.refinement|var2.refinement),
                          sigma=(self.sigma**2+var2.sigma**2)**0.5
                          )
        else:
            res = self.value-var2
        return res
    def __rsub__(self, var2):
        if type(var2) is numpy.ndarray:
            res_1 = var2-self.value
            res = res_1.astype(var2.dtype)
        elif isinstance(var2, Fitable):
            res = Fitable(value=var2.value-self.value,
                          refinement=(self.refinement|var2.refinement),
                          sigma=(self.sigma**2+var2.sigma**2)**0.5
                          )
        else:
            res = var2-self.value
        return res
    def __mul__(self, var2):
        if type(var2) is numpy.ndarray:
            res_1 = self.value*var2
            res = res_1.astype(var2.dtype)
        elif isinstance(var2, Fitable):
            res = Fitable(value=var2.value*self.value,
                          refinement=(self.refinement|var2.refinement),
                          sigma=((self.sigma*var2.value)**2+(var2.sigma*self.value)**2)**0.5
                          )
        else:
            res = self.value*var2
        return res 
    def __rmul__(self, var2):
        if type(var2) is numpy.ndarray:
            res_1 = self.value*var2
            res = res_1.astype(var2.dtype)
        elif isinstance(var2, Fitable):
            res = Fitable(value=var2.value*self.value,
                          refinement=(self.refinement|var2.refinement),
                          sigma=((self.sigma*var2.value)**2+(var2.sigma*self.value)**2)**0.5
                          )
        else:
            res = self.value*var2
        return res
    def __div__(self, var2):
        if type(var2) is numpy.ndarray:
            res_1 = self.value*1./var2
            res = res_1.astype(var2.dtype)
        elif isinstance(var2, Fitable):
            hh = (((var2.value*self.sigma)**2+(self.value*var2.sigma)**2)**0.5)/(var2.value)**2
            res = Fitable(value=self.value*1./var2.value, 
                          refinement=(self.refinement | var2.refinement),
                          sigma=hh)
        else:
            res = self.value*1./var2
        return res 
    def __rdiv__(self, var2):
        if type(var2) is numpy.ndarray:
            res_1 = var2*1./self.value
            res = res_1.astype(var2.dtype)
        elif isinstance(var2, Fitable):
            hh = (((var2.value*self.sigma)**2+(self.value*var2.sigma)**2)**0.5)/(self.value)**2
            res = Fitable(value=var2.value*1./self.value, 
                          refinement=(self.refinement | var2.refinement),
                          sigma=hh)
        else:
            res = var2*1./self.value
        return res 
    def __truediv__(self, var2):
        if type(var2) is numpy.ndarray:
            res_1 = self.value*1./var2
            res = res_1.astype(var2.dtype)
        elif isinstance(var2, Fitable):
            hh = (((var2.value*self.sigma)**2+(self.value*var2.sigma)**2)**0.5)/(var2.value)**2
            res = Fitable(value=self.value*1./var2.value, 
                          refinement=(self.refinement | var2.refinement),
                          sigma=hh)
        else:
            res = self.value*1./var2
        return res 
    def __rtruediv__(self, var2):
        if type(var2) is numpy.ndarray:
            res_1 = var2*1./self.value
            res = res_1.astype(var2.dtype)
        elif isinstance(var2, Fitable):
            hh = (((var2.value*self.sigma)**2+(self.value*var2.sigma)**2)**0.5)/(self.value)**2
            res = Fitable(value=var2.value*1./self.value, 
                          refinement=(self.refinement | var2.refinement),
                          sigma=hh)
        else:
            res = var2*1./self.value
        return res 
    def exp(self):
        return exp(self.value)
    def __pos__(self):
        return self.value
    def __neg__(self):
        return -1*self.value
    def __abs__(self):
        return abs(self.value)
    def __round__(self, n):
        return round(self.value, n)
    def __pow__(self, var2):
        return self.value**var2
    def __rpow__(self, var2):
        return var2**self.value
    def __mod__(self, var2):
        return self.value%var2
    def __rmod__(self, var2):
        return var2%self.value
    def __floordiv__(self, var2):
        return self.value//var2
    def __rfloordiv__(self, var2):
        return var2//self.value
    def __ne__(self, var2):
        return  self.value != var2    
    def __rne__(self, var2):
        return  var2 != self.value  
    def __gt__(self, var2):
        return self.value > var2
    def __ge__(self, var2):
        return self.value >= var2
    def __lt__(self, var2):
        return  self.value  < var2
    def __le__(self, var2):
        return  self.value  <= var2
    def __and__(self, var2):
        return  (self.refinement & var2)
    def __rand__(self, var2):
        return  (var2 & self.refinement)
    def __or__(self, var2):
        return  (self.refinement | var2)
    def __ror__(self, var2):
        return  (var2 | self.refinement)




"""
CrysPy
======

To run rhochi refinement from a command-line:
>>> python -m cryspy.run_rhochi

To run a simple gui interface for rhochi from a command-line:
>>> python -m cryspy.gui_rhochi

To create a template in the folder run from a command-line:
>>> python -m cryspy.rhochi_diffrn # template for single-crystal diffraction experiment
>>> python -m cryspy.rhochi_pd # template for 1d powder diffraction experiment
>>> python -m cryspy.rhochi_pd2d # template for 2d powder diffraction experiment
>>> python -m cryspy.rhochi_pd2dt # template for 2d powder diffraction with one-axial texture model experiment

Objects to describe crystal structure
---------------------------------------
- Crystal
    the main object to describe crystal structure
    - SpaceGroup
    - Cell
    - AtomType
    - AtomSite
    - AtomSiteMagnetism
    - AtomSiteMagnetismAniso
    - AtomSiteAniso

Objects to describe experiments
---------------------------------------
- BeamPolarization
    common object for all polaraized neutron diffraction experiments
- Diffrn
    the main object to describe single crystal diffraction experiment
    - OrientMatrix 
    - Extinction 
    - DiffrnRefln 
- Pd
    the main object to describe 1d polaraized neutrhon powder diffraction experiment
    - PdMeas
        measurements
    - PdBackground
    - PdExclude2Theta
    - PdInstrReflexAsymmetry
    - PdInstrResolution
    - PdPhase
    - PdPeak
    - PdProc
- Pd2d
    the main object to describe 2d polaraized neutrhon powder diffraction experiment
    - Pd2dMeas
        measurements
    - Pd2dBackground
    - Pd2dExclude2Theta
    - Pd2dInstrReflexAsymmetry
    - Pd2dInstrResolution
    - Pd2dPhase
    - Pd2dPeak
    - Pd2dProc
- Pd2dt
    the main object to describe 2d polaraized neutrhon powder diffraction experiment in case of one-axial texture
    
Use help function to get information for each of them:
>>> import cryspy
>>> help(cryspy.Cell)
    class Cell(builtins.object)
     |  Cell(a=1.0, b=1.0, c=1.0, alpha=90.0, beta=90.0, gamma=90.0, bravais_lattice='triclinic')
     |
     |  Data items in the Cell class record details about the
     |  crystallographic cell parameters and their measurement.
     ...
     
Specify the input parameters of the object to get detailed description
>>> help(cryspy.Cell.bravais_lattice)
"""
name = "cryspy"

from .f_common.cl_fitable import Fitable

from .scripts.rhochi.cl_rhochi import RhoChi, rhochi_refinement, rhochi_read_file


from .f_crystal import (
    SpaceGroup,
    Cell,
    AtomType,
    AtomSite,
    Crystal,
    calc_mRmCmRT,
    AtomSiteMagnetism,
    AtomSiteMagnetismAniso,
    AtomSiteAniso
    )

from .f_experiment import (
    BeamPolarization,
    OrientMatrix,
    Extinction,
    DiffrnRefln, 
    Diffrn,
    PdBackground,
    PdExclude2Theta,
    PdInstrReflexAsymmetry,
    PdInstrResolution,
    PdMeas,
    PdPeak,
    PdPhase,
    PdProc,
    Pd,
    Pd2dBackground,
    Pd2dExclude,
    Pd2dInstrReflexAsymmetry,
    Pd2dInstrResolution,
    Pd2dMeas,
    Pd2dPhase,
    Pd2dPeak,
    Pd2dProc,
    Pd2d,
    Pd2dt
    )

#from .scripts.rhochi.rhochi_viewer import main as rhochi_gui

def crystal_from_file(f_name):
    rhochi = RhoChi()
    with open(f_name, 'r') as fid:
        string = fid.read()
    rhochi.from_cif(string)
    if len(rhochi.crystals) != 0:
        crystal = rhochi.crystals[0]
    else:
        crystal = Crystal()
        print("Information about crystal is not found in the file: '{:}'".format(f_name))
    return crystal




"""
define class Pd2dBackground which describes the background in 2d powder diffraction experiment
"""
__author__ = 'ikibalin'
__version__ = "2019_09_10"
import os
import numpy
import scipy
import scipy.interpolate


from pycifstar import Global
from cryspy.f_common.cl_fitable import Fitable


class Pd2dBackground(object):
    """
    Example:

    _pd2d_background_2theta_phi_intensity
    ;
         2    4.5     40.0     80.0
    -3.000 -350.0   -350.0   -400.0
    41.000 -350.0   -350.0   -400.0
    ;
    """
    def __init__(self, ttheta=[], phi=[], intensity=[[]]):
        super(Pd2dBackground, self).__init__()
        self.__pd2d_background_phi = []
        self.__pd2d_background_2theta = []
        self.__pd2d_background_2theta_phi_intensity = [[]]
        self.phi = phi
        self.ttheta = ttheta
        self.intensity = intensity

    @property
    def phi(self):
        return tuple(self.__pd2d_background_phi)
    @phi.setter
    def phi(self, l_x):
        l_x_in = []
        for x in l_x:
            x_in = float(x)
            l_x_in.append(x_in)
        self.__pd2d_background_phi = l_x_in
        len_x = len(l_x_in)

        for ind, l_intensity in enumerate(self.intensity):
            len_1 = len(l_intensity)
            if len_1 > len_x:
                self.__pd2d_background_2theta_phi_intensity[ind] = self.__pd2d_background_2theta_phi_intensity[ind][:len_x]
            elif len_1 < len_x:
                l_fitable = [Fitable(value=0., name="_pd2d_background_2theta_phi_intensity") for hh in range(len_x-len_1)] #default
                self.__pd2d_background_2theta_phi_intensity[ind].extend(l_fitable)




    @property
    def ttheta(self):
        return tuple(self.__pd2d_background_2theta)
    @ttheta.setter
    def ttheta(self, l_x):
        l_x_in = []
        for x in l_x:
            x_in = float(x)
            l_x_in.append(x_in)
        self.__pd2d_background_2theta = l_x_in

        len_x = len(l_x_in)
        len_1 = len(self.intensity)
        if len_1 > len_x:
            self.__pd2d_background_2theta_phi_intensity = self.__pd2d_background_2theta_phi_intensity[:len_x]
        elif len_1 < len_x:
            l_fitable = [[] for hh in range(len_x-len_1)] #default
            self.__pd2d_background_2theta_phi_intensity.extend(l_fitable)        

    @property
    def intensity(self):
        return tuple(self.__pd2d_background_2theta_phi_intensity)
    @intensity.setter
    def intensity(self, ll_x):
        ll_fitable = []
        for l_x in ll_x:
            l_fitable = []
            for x in l_x:
                if isinstance(x, Fitable):
                    x_in = x
                else:
                    x_in = Fitable()
                    x_in.name = "_pd2d_background_2theta_phi_intensity"
                    flag = x_in.take_it(x)
                l_fitable.append(x_in)
            ll_fitable.append(l_fitable)
        len_x = len(ll_fitable)

        len_1 = len(self.ttheta)
        if len_1 < len_x:
            ll_fitable = ll_fitable[:len_1]
        elif len_1 > len_x:
            ll_fitable.extend([[] for hh in range(len_1-len_x)])

        len_1 = len(self.phi)
        for l_fitable in ll_fitable:
            len_x = len(l_fitable)
            if len_1 < len_x:
                l_fitable = l_fitable[:len_1]
            elif len_1 > len_x:
                l_fitable.extend([Fitable(value=0., name="_pd2d_background_2theta_phi_intensity") for hh in range(len_1-len_x)])

        self.__pd2d_background_2theta_phi_intensity = ll_fitable



            
    def __repr__(self):
        ls_out = ["Pd2dBackground:"]
        ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
        ll_intensity = self.intensity
        ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
        for phi, l_intensity in zip(self.phi, ll_intensity):
            ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_.print_with_sigma) for _ in l_intensity]))
        return "\n".join(ls_out)

    @property
    def to_cif(self):
        ls_out = []
        if self.is_defined:
            ls_out.append("_pd2d_background_2theta_phi_intensity")
            ls_out.append(";")
            ls_out.append("{:12} ".format(len(self.phi)) + " ".join(["{:6.2f}      ".format(_) for _ in self.ttheta]))
            ll_intensity = self.intensity
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            for phi, l_intensity in zip(self.phi, ll_intensity):
                ls_out.append("{:12.2f} ".format(phi) + " ".join(["{:12}".format(_.print_with_sigma) for _ in l_intensity]))
            ls_out.append(";")
        return "\n".join(ls_out)

    def from_cif(self, string: str):
        cif_global = Global()
        flag = cif_global.take_from_string(string)
        if not flag:
            return False
        flag = False
        flag = cif_global.is_prefix("_pd2d_background_2theta_phi_intensity")
        if flag:
            cif_value = cif_global["_pd2d_background_2theta_phi_intensity"]
            string = cif_value.value
            l_1 = string.strip().split("\n")
            l_ttheta = [float(_) for _ in l_1[0].strip().split()[1:]]
            l_phi, ll_intensity = [], []
            for line in l_1[1:]:
                l_1 = line.strip().split()
                l_phi.append(float(l_1[0]))
                ll_intensity.append(l_1[1:])
            ll_intensity = [[ll_intensity[_2][_1] for _2 in range(len(ll_intensity))] for _1 in range(len(ll_intensity[0]))]
            self.phi = l_phi
            self.ttheta = l_ttheta
            self.intensity = ll_intensity
        else:
            self.phi, self.ttheta, self.intensity = [], [], [[]]
        return True

    @property
    def is_defined(self):
        cond = all([self.phi is not None, self.ttheta is not None, self.intensity is not None])
        return cond

    @property
    def is_variable(self):
        res = any([[_2.refinement for _2 in _1] for _1 in self.intensity])
        return res
    
    def get_variables(self):
        l_variable = []
        for _1 in self.intensity:
            for _2 in _1:
                if _2.refinement:
                    l_variable.append(_2)
        return l_variable

    def _show_message(self, s_out: str):
        print("***  Error ***")
        print(s_out)

    def interpolate_by_points(self, tth, phi):
        l_phi_b = self.phi
        l_tth_b = self.ttheta
        ll_int_b = self.intensity
        ll_int_b = [[float(ll_int_b[_2][_1]) for _2 in range(len(ll_int_b))] for _1 in range(len(ll_int_b[0]))]
        if len(l_tth_b) == 0:
            int_2d = numpy.zeros((tth.size, phi.size), dtype=float)
        else:
            phi_b = numpy.array(l_phi_b, dtype=float)
            tth_b = numpy.array(l_tth_b, dtype=float)
            int_b = numpy.array(ll_int_b, dtype=float)
            func = scipy.interpolate.interp2d(tth_b, phi_b, int_b)
            #tth_2d, phi_2d = numpy.meshgrid(tth, phi, indexing="ij")
            int_2d = func(tth, phi)
            int_2d = int_2d.transpose()
        return int_2d




__version__ = '1.0.0'
__author__ = 'Iurii Kibalin'
__contact__ = 'iurii.kibalin@cea.fr'

from cryspy.f_experiment.f_single_domain.cl_experiment_single_domain import ExperimentSingleDomain
from cryspy.f_experiment.f_single_domain.cl_observed_data_single_domain import ObservedDataSingleDomain
